#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Automate generation of parametrized launchers."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2016, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import contextlib
import functools
import operator
import os
import shutil
import six
import subprocess
import warnings

from sciexp2.common import parallel as parallel_pkg
from sciexp2.common import progress
from sciexp2.common import utils
from sciexp2.common import varref
import sciexp2.common.instance
import sciexp2.common.filter
from sciexp2.common import pp
from sciexp2 import templates
from sciexp2 import launcher


_LISTXLATOR = varref.ListXlator("@")
_DICTXLATOR = varref.DictXlator("@")
_EXPANDER = sciexp2.common.instance.Expander("@")


######################################################################
# Helper functions

def file_contents(path, mode="r"):
    """Return the contents of given path as a string.

    Can be used as an argument to any method in `Launchgen` that accepts an
    expression. If the path itself is an expression, the appropriate file will
    be read after expanding its path expression.

    """
    variables = varref.expr_get_vars("@", path)

    def read_file(path, mode):
        with open(path, mode) as f:
            return "".join(f.readlines())

    if len(variables) == 0:
        return read_file(path, mode)

    def deferred_read_file(*values):
        real_path = _LISTXLATOR.xlate(path, [str(v) for v in values])
        return read_file(real_path, mode)

    return defer(deferred_read_file, *[defer(v) for v in variables])


######################################################################
# Partial views (function identification)

VIEW_SUPPORTED = {}


def _supports_view(other):
    assert callable(other)

    def _func_decorator(func):
        assert callable(func)
        assert func not in VIEW_SUPPORTED
        VIEW_SUPPORTED[func.__name__] = other.__name__
        return func

    return _func_decorator


######################################################################
# Launchgen

class Launchgen (pp.Pretty):
    """Define and generate program launchers.

    Parameters
    ----------
    contents : optional
        Initial contents.
    out : optional
        Initial value for the `out` attribute.
    dereference : bool, optional
        Initial value for the `dereference` attribute.

    Attributes
    ----------
    out : str
        Output directory for all non-absolute paths.
    contents : `~sciexp2.common.instance.InstanceGroup`
        Instances describing the elements of this dimension.

    See also
    --------
    file_contents

    """

    #: Output directory for all non-absolute paths.
    out = "./out"

    #: Whether to dereference source symlinks during pack.
    dereference = False

    def __init__(self, contents=None, **kwargs):
        out = kwargs.pop("out", None)
        dereference = kwargs.pop("dereference", None)
        utils.assert_kwargs(kwargs)

        if contents is None:
            contents = sciexp2.common.instance.InstanceGroup()
            # add dummy contents
            contents += sciexp2.common.instance.Instance()
        if isinstance(contents, Launchgen) or\
           isinstance(contents, LaunchgenView):
            if out is None:
                out = contents.out
            contents = contents.contents
            if dereference is None:
                dereference = contents.dereference
        if out is not None:
            self.out = out
        if dereference is not None:
            self.dereference = dereference

        self._contents = sciexp2.common.instance.InstanceGroup(contents)

        self._deferred = set()

    @property
    def contents(self):
        """Instances describing the elements of this dimension."""
        return self._contents

    def _do_repr_pretty_(self, select, p, cycle):
        if select is None:
            select = sciexp2.common.filter.Filter()
        with self.pformat(p, cycle):
            if len(list(self._contents.variables())) > 0:
                p.pretty(list(self._contents.select(select)))
                p.text(",")
                p.breakable()
            p.text("out=%r" % self.out)

    @_supports_view(_do_repr_pretty_)
    def _repr_pretty_(self, p, cycle):
        self._do_repr_pretty_(None, p, cycle)

    def select(self, *filters):
        """Get a `LaunchgenView` to operate on the subset matching `filters`.

        Notes
        -----
        The relative order inside `contents` between new and old instances is
        mantained, such that the result appears as if instances matching the
        filters were "expanded" in-place.

        Examples
        --------
        All operations are performed only on selected elements:

        >>> l = Launchgen()
        >>> l.params(a=range(2), b=range(2))
        >>> l
        Launchgen([Instance({'a': 0, 'b': 0}),
                   Instance({'a': 0, 'b': 1}),
                   Instance({'a': 1, 'b': 0}),
                   Instance({'a': 1, 'b': 1})],
                  out='./out')
        >>> v = l.select("a == b")
        >>> v.params(c=(defer("a") + defer("b")))
        >>> v
        Launchgen([Instance({'a': 0, 'b': 0, 'c': 0}),
                   Instance({'a': 1, 'b': 1, 'c': 2})],
                  out='./out')

        And appear on the original object as if they happened "in-place":

        >>> l
        Launchgen([Instance({'a': 0, 'b': 0, 'c': 0}),
                   Instance({'a': 0, 'b': 1}),
                   Instance({'a': 1, 'b': 0}),
                   Instance({'a': 1, 'b': 1, 'c': 2})],
                  out='./out')

        The inverse selection is also readily available:

        >>> i = v.select_inverse()
        >>> i
        Launchgen([Instance({'a': 0, 'b': 1}), Instance({'a': 1, 'b': 0})],
                  out='./out')
        >>> i.params(d=(defer("a") * defer("b")))
        >>> l
        Launchgen([Instance({'a': 0, 'b': 0, 'c': 0}),
                   Instance({'a': 0, 'b': 1, 'd': 0}),
                   Instance({'a': 1, 'b': 0, 'd': 0}),
                   Instance({'a': 1, 'b': 1, 'c': 2})],
                  out='./out')

        As well as nested selections are also supported:

        >>> n = v.select("c == 0")
        >>> n.params(e=[10, 20])
        >>> n.select_inverse().params(e=[30, 40])
        >>> l
        Launchgen([Instance({'a': 0, 'b': 0, 'c': 0, 'e': 10}),
                   Instance({'a': 0, 'b': 0, 'c': 0, 'e': 20}),
                   Instance({'a': 0, 'b': 1, 'd': 0}),
                   Instance({'a': 1, 'b': 0, 'd': 0}),
                   Instance({'a': 1, 'b': 1, 'c': 2, 'e': 30}),
                   Instance({'a': 1, 'b': 1, 'c': 2, 'e': 40})],
                  out='./out')

        """
        return LaunchgenView(self,
                             sciexp2.common.filter.Filter("True"),
                             sciexp2.common.filter.Filter(*filters))

    #: Default standard input for `execute`
    STDIN = None
    #: Default standard output for `execute`
    STDOUT = None
    #: Default standard error for `execute`
    STDERR = None

    def _execute(self, select, *args, **kwargs):
        parallel = kwargs.pop("parallel", None)
        std_args = {
            'stdin':  kwargs.pop("stdin",  self.STDIN),
            'stdout': kwargs.pop("stdout", self.STDOUT),
            'stderr': kwargs.pop("stderr", self.STDERR),
        }
        utils.assert_kwargs(kwargs)

        def std_needs_translate(name):
            arg = std_args[name]
            return isinstance(arg, six.string_types) and \
                varref.expr_get_vars("@", arg, accept_regexp=True)
        std_vars = {
            'stdin':  std_needs_translate('stdin'),
            'stdout': std_needs_translate('stdout'),
            'stderr': std_needs_translate('stderr'),
        }

        if len(args) == 0:
            return

        cmd_arg_sep = "|;|"
        cmd = cmd_arg_sep.join(args)

        expansions = _EXPANDER.expand(cmd, self._contents,
                                      filter_=select,
                                      recursive=True)

        @contextlib.contextmanager
        def get_stdfile_null():
            yield None

        def get_stdfile(name, expansion):
            arg = std_args[name]
            if std_vars[name]:
                instance = expansion.instances.get_index(0)
                path = utils.get_path(_DICTXLATOR.xlate_rec(arg,
                                                            instance))
                return open(path, "w")
            elif isinstance(arg, six.string_types):
                path = utils.get_path(arg)
                return open(path, "w")
            elif arg is None:
                return get_stdfile_null()
            else:
                return arg

        def do_execute(ecmd):
            real_cmd = ecmd.text
            if len(args) == 1:
                progress.log(progress.LVL_VERBOSE,
                             "", real_cmd, "\n")
                real_cmd = [real_cmd]
            else:
                real_cmd = [real_cmd.split(cmd_arg_sep)]
                progress.log(progress.LVL_VERBOSE,
                             "", " ".join(real_cmd[0]), "\n")

            with get_stdfile("stdin",  ecmd) as stdin,\
                 get_stdfile("stdout", ecmd) as stdout,\
                 get_stdfile("stderr", ecmd) as stderr:
                res = subprocess.call(*real_cmd,
                                      shell=len(args) == 1,
                                      stdin=stdin,
                                      stdout=stdout,
                                      stderr=stderr)
            if res != 0:
                raise Exception("Command did not finish correctly: "
                                "%s -> %d" % (ecmd.text, res))

        with progress.get(expansions, msg="Executing commands...") as progr:
            if parallel is False:
                for ecmd in six.itervalues(expansions):
                    do_execute(ecmd)
                    progr()
            else:
                if parallel is True:
                    parallel = None
                work = list(six.itervalues(expansions))
                for i in parallel_pkg.t_imap_unordered(
                        work, do_execute, parallelism=parallel):
                    progr()

    @_supports_view(_execute)
    def execute(self, *args, **kwargs):
        """Execute an external program of shell command.

        If a single string is passed, it's interpreted as an expression that is
        expanded and then executed as a shell command.

        A list of strings is interpreted as a program name and its list of
        arguments (all treated as expressions and expanded).

        Parameters
        ----------
        args : str or list of str
            Command to execute.
        stdin, stdout, stderr : file or str, optional
            Paths for the standard input/output/error for the command (defaults
            to `STDIN`/`STDOUT`/`STDERR`).
        parallel : optional
            Whether to execute commands in parallel using separate processes
            (default is `False`).

        Notes
        -----
        If strings are used for `stdin`, `stdout` or `stderr`, these are
        treated as expressions and, thus, are translated accordingly.

        Command execution is internally handled by `subproces.call` and thus
        the `stdin`, `stdout` and `stderr` arguments can also have any value
        accepted by it.

        See also
        --------
        sciexp2.common.parallel.get_parallelism

        """
        self._execute(None, *args, **kwargs)

    #: Variables added by `_find_files_generic`
    _find_files_generic__VARS = set(["FILE"])

    def _find_files_generic(self, select, expression, filters,
                            variables, progr_msg,
                            process=None):
        if process is None:
            def generic_process(instance):
                yield instance
            process = generic_process

        if select is not None:
            selected_indexes = [idx for idx, i in enumerate(self._contents)
                                if select.match(i)]
            if len(selected_indexes) == 0:
                contents = sciexp2.common.instance.InstanceGroup(
                    self._contents)
                contents_rest = sciexp2.common.instance.InstanceGroup()
            else:
                last = max(selected_indexes)
                contents = sciexp2.common.instance.InstanceGroup(
                    self._contents.get_index(slice(0, last+1)))
                contents_rest = sciexp2.common.instance.InstanceGroup(
                    self._contents.get_index(slice(last+1, -1)))
        else:
            contents = sciexp2.common.instance.InstanceGroup(self._contents)
            contents_rest = sciexp2.common.instance.InstanceGroup()

        # take care of dummy contents
        if len(contents) is 1 and len(list(contents.variables())) is 0:
            contents = sciexp2.common.instance.InstanceGroup()

        filters = sciexp2.common.filter.Filter(*filters)
        fvariables = set(varref.expr_get_vars("@", expression, True)) |\
                     self._find_files_generic__VARS | variables
        filters.validate(fvariables)

        files = utils.find_files("@", expression, "FILE")
        with progress.progressable_simple(
                files, None, msg=progr_msg) as pfiles:
            for instance in pfiles:
                for instance in process(instance):
                    if filters.match(instance):
                        contents.add(instance)

        contents = contents + contents_rest

        # re-add dummy contents, if necessary
        if len(contents) == 0:
            contents += sciexp2.common.instance.Instance()

        self._contents = contents

    def _find_files(self, select, expression, *filters):
        return self._find_files_generic(select, expression, filters, set(),
                                        "finding files...")

    @_supports_view(_find_files)
    def find_files(self, expression, *filters):
        """Find paths matching a regular expression and add to `contents`.

        Parameters
        ----------
        expression : expression
            An expression for selecting paths.
        filters : list of filters, optional
            Filters that paths must match to be considered. All provided
            filters are AND'ed.

        Notes
        -----
        For each path matching `expression` and `filters`, extract it into an
        `~sciexp2.common.instance.Instance` and add to `contents`.

        The expression can contain arbitrarily complex `regular expressions
        <http://docs.python.org/library/re.html#regular-expression-syntax>`_,
        and variables are extracted from:

            - variable references: ``@varname@``
            - named re groups: ``(?P<varname>regexp)``

        Implicitly added variables:
            ======== ======================
            Name     Contents
            ======== ======================
            ``FILE`` the complete path name
            ======== ======================

        See also
        --------
        find_SPEC, find_simpoint

        """
        return self._find_files(None, expression, *filters)

    _find_SPEC__VARS = set(["SET", "BENCHMARK", "TEST", "DIR", "BINARY",
                            "STDIN", "PARAMS"])

    def _find_SPEC(self, select, base_dir, *filters):
        expression = os.path.join(base_dir, "benchspec",
                                  "@SET@", "@BENCHMARK@",
                                  "run", "(?:run_base_|000).*",
                                  "speccmds.cmd$")
        if not os.path.isabs(expression):
            expression = "./" + expression

        def process(instance):
            with open(utils.get_path(instance["FILE"])) as f:
                lines = f.readlines()

                # support both SPECcpu2006 and SPECcpu2000 style (respectively)
                assert lines[0].startswith("-C ") or lines[0].startswith("-u")

                dir_path = lines[0].split()[1]

                test = 0
                for line in lines[1:]:
                    new_instance = sciexp2.common.instance.Instance(instance)

                    elems = line.split()
                    if elems[0] == "-i":      # stdin
                        new_instance["STDIN"] = elems[1]
                        elems = elems[2:]
                    if elems[0] == "-o":      # stdout
                        elems = elems[2:]
                    if elems[0] == "-e":      # stderr
                        elems = elems[2:]
                    new_instance["DIR"] = dir_path
                    new_instance["BINARY"] = os.path.basename(elems[0])
                    new_instance["TEST"] = test
                    new_instance["PARAMS"] = " ".join(elems[1:])
                    new_instance["FILE"] = dir_path + os.sep +\
                                           new_instance["BINARY"]

                    test += 1
                    yield new_instance

        return self._find_files_generic(select, expression, filters,
                                        self._find_SPEC__VARS,
                                        "finding SPEC benchmarks...",
                                        process)

    @_supports_view(_find_SPEC)
    def find_SPEC(self, base_dir, *filters):
        """Find SPEC benchmarks and extract parameters.

        First searches the ``speccmds.cmd`` files for each benchmark with
        ``base_dir +
        "/benchspec/@SET@/@BENCHMARK@/run/(?:run_base_|000).*/speccmds.cmd"``,
        and then parses the ``speccmds.cmd`` files in order to generate one
        instance for each benchmark execution test.

        Parameters
        -----------
        base_dir : expression
            SPEC's base directory
        filters : list of filters, optional
            Filters that paths must match to be considered. All provided
            filters are AND'ed.

        Notes
        -----
        Implicitly added variables:
            ============= ======================
            Name          Contents
            ============= ======================
            ``SET``       set (e.g., CINT2000, CFP2000, CPU2006)
            ``BENCHMARK`` name (e.g., 400.perlbench)
            ``TEST``      test number
            ``DIR``       absolute path to prepared run directory
            ``BINARY``    executable file name
            ``STDIN``     standrad input file name
            ``PARAMS``    command line parameters
            ``FILE``      absolute path to executable
            ============= ======================

        See also
        --------
        find_files

        References
        ----------
        .. [1] http://www.spec.org

        """
        return self._find_SPEC(None, base_dir, *filters)

    def _find_simpoint(self, select, simpoints_expr, weights_expr, ignore):
        contents = self._contents

        # take care of dummy contents
        if len(contents) is 1 and len(list(contents.variables())) is 0:
            contents = sciexp2.common.instance.InstanceGroup()

        if select is None:
            select = sciexp2.common.filter.Filter()

        new_contents = sciexp2.common.instance.InstanceGroup()

        with progress.progressable_simple(
                contents, None, msg="finding simpoints...") as pcontents:
            for instance in pcontents:
                if not select.match(instance):
                    continue

                simpoints = _DICTXLATOR.xlate_rec(simpoints_expr, instance)
                simpoints = utils.get_path(simpoints)
                if not os.path.isfile(simpoints):
                    if not ignore:
                        raise ValueError("Could not find simpoints file: " +
                                         simpoints)
                    continue
                weights = _DICTXLATOR.xlate_rec(weights_expr, instance)
                weights = utils.get_path(weights)
                if not os.path.isfile(weights):
                    if not ignore:
                        raise ValueError("Could not find "
                                         "simpoint weights file: " + weights)
                    continue

                simpoints_list = []
                for line in open(simpoints):
                    (sp, idx) = line.split()
                    simpoints_list.insert(int(idx), sp)

                weights_list = []
                for line in open(weights):
                    (wg, idx) = line.split()
                    weights_list.insert(int(idx), wg)

                assert len(simpoints_list) == len(weights_list)
                for idx in range(0, len(simpoints_list)):
                    new_instance = sciexp2.common.instance.Instance(instance)
                    new_instance["SIMPOINT_ID"] = idx
                    new_instance["SIMPOINT_SP"] = simpoints_list[idx]
                    new_instance["SIMPOINT_WG"] = weights_list[idx]

                    new_contents.add(new_instance)

        # re-add dummy contents, if necessary
        if len(new_contents) == 0:
            new_contents += sciexp2.common.instance.Instance()

        self._contents = new_contents

    @_supports_view(_find_simpoint)
    def find_simpoint(self, simpoints_expr, weights_expr, ignore=False):
        """Expand current contents with SimPoint information.

        Expands the given arguments with current `contents` to find the
        results of `SimPoint` executions. This is useful to generate launchers
        for simulating what `SimPoint` found as representative.

        Parameters
        ----------
        simpoints_expr, weights_expr : expression
            Expression SimPoint's ``simpoints`` and ``weights`` output files.
        ignore : bool, optional
            Whether to ignore current `contents` that have no simpoint
            information available; otherwise aborts execution.

        Notes
        -----
        Implicitly added variables:
            =============== ======================
            Name            Contents
            =============== ======================
            ``SIMPOINT_ID`` simpoint index
            ``SIMPOINT_SP`` simpoint (e.g., basic block number)
            ``SIMPOINT_WG`` simpoint weight
            =============== ======================

        See also
        --------
        find_files

        References
        ----------
        .. [1] http://cseweb.ucsd.edu/~calder/simpoint

        """
        return self._find_simpoint(None, simpoints_expr, weights_expr, ignore)

    def _params(self, select, *filters, **params):
        append = params.pop("append", False)
        assert isinstance(append, bool)

        self._contents, self._deferred = _do_params(select, self._contents,
                                                    self._deferred, filters,
                                                    params, append)

    @_supports_view(_params)
    def params(self, *filters, **params):
        """Expand current `contents` with all valid parameter permutations.

        Substitutes the current `contents` with the `cartesian product` of
        current `contents` and all given parameter permutations.

        Parameters
        ----------
        filters : list of filters, optional
            Discard results not matching the given filters (default is to
            accept all).
        params : dict
            A dictionary of variable name / values pairs to perform
            permutations.

        Notes
        -----
        Values in `params` can have two forms:

         - A single value::

            params(a = 1)

         - An iterable container (e.g., a list to permute with all other
           variables, including the parameters)::

            params(a = [ 1, 10 ])

        Any of these forms can contain any of the following types (which can be
        mixed):

         - An immediate::

            params(var1 = [ 1, "valueN", math.log(1024, 2), ... ])

         - An :term:`expression` that will be :term:`expanded <expansion>`::

            params(var1 = [ "@other_param@-@another_one@", ...])

         - Deferred values:

           See `defer` for deferring the evaluation of still undefined
           variables.

        Providing a variable in `params` that already exists in `contents` will
        override its value. Note that the new value can reference any other
        variable in `contents`. Also, if the new value references variables not
        present in `contents`, values from `params` will be used instead.

        If variable `append` is present and set to `True`, the parameters will
        be appended to the contents, instead of being recombined with current
        contents.

        Examples
        --------
        >>> l = Launchgen()
        >>> l.params("a % 2 == 0",
        ...          a = range(4),
        ...          b = [ "foo", "bar" ])
        >>> l
        Launchgen([Instance({'a': 0, 'b': 'foo'}),
                   Instance({'a': 0, 'b': 'bar'}),
                   Instance({'a': 2, 'b': 'foo'}),
                   Instance({'a': 2, 'b': 'bar'})],
                  out='./out')

        New values reference old values (when available):

        >>> l.params(b = "@b@ (bis)",
        ...          c = "@b@-@a@")
        >>> l
        Launchgen([Instance({'a': 0, 'b': 'foo (bis)', 'c': 'foo-0'}),
                   Instance({'a': 0, 'b': 'bar (bis)', 'c': 'bar-0'}),
                   Instance({'a': 2, 'b': 'foo (bis)', 'c': 'foo-2'}),
                   Instance({'a': 2, 'b': 'bar (bis)', 'c': 'bar-2'})],
                  out='./out')

        New parameter combinations can also be appended, instead of combined
        with the existing ones:

        >>> with l.select("a == 0") as s:
        ...     s.params(append = True, x=[100, 200])
        >>> l
        Launchgen([Instance({'a': 0, 'b': 'foo (bis)', 'c': 'foo-0'}),
                   Instance({'a': 0, 'b': 'bar (bis)', 'c': 'bar-0'}),
                   Instance({'x': 100}),
                   Instance({'x': 200}),
                   Instance({'a': 2, 'b': 'foo (bis)', 'c': 'foo-2'}),
                   Instance({'a': 2, 'b': 'bar (bis)', 'c': 'bar-2'})],
                  out='./out')

        See also
        --------
        defer, sciexp2.common.filter.PFilter

        """
        return self._params(None, *filters, **params)

    def _pack(self, select, from_expr, to_expr, dereference=None):
        if dereference is None:
            dereference = self.dereference
        done = {}
        done_instances = {}
        expansions = _EXPANDER.expand(from_expr, self._contents,
                                      filter_=select,
                                      recursive=True)

        with progress.get(self._contents, msg="copying files...") as progr:
            for exp in six.itervalues(expansions):
                for instance in exp.instances:
                    progr()

                    exp_from = utils.get_path(exp.text)
                    exp_to = _DICTXLATOR.xlate_rec(to_expr, instance)
                    exp_to = utils.get_path(exp_to)

                    idx = exp_from + "+" + exp_to
                    if idx in done:
                        if not done[idx]:
                            done[idx] = True
                            progress.log(
                                progress.LVL_INFO,
                                "Skipping already packed file:",
                                exp_from, "->", exp_to, "\n")
                            progress.log(
                                progress.LVL_VERBOSE,
                                "Conflicting instances:\n",
                                done_instances[idx], "\n", instance, "\n")
                        continue
                    done[idx] = False
                    done_instances[idx] = instance

                    progress.log(progress.LVL_DEBUG,
                                 "Packing:", exp_from, "->", exp_to, "\n")

                    exp_to = self.out + os.sep + exp_to
                    utils.copy_path(exp_from, exp_to)

    @_supports_view(_pack)
    def pack(self, from_expr, to_expr, dereference=None):
        """Copy files from source to destination.

        Parameters
        ----------
        from_expr, to_expr : expression
            Source/destination file expressions.
        dereference : bool, optional
            Whether to dereference symlinks in source files (defaults to
            ``self.dereference``).

        Notes
        -----
        Multiple source expressions could have a single destination, in which
        case this is noted with the message ``Skipping already packed file:
        <source> -> <destination>``.

        The copy is performed only if the source is newer than the destination,
        or if the destination does not exist.

        """
        return self._pack(None, from_expr, to_expr, dereference)

    def _translate(self, select, expression, instance=None):
        if instance:
            return _DICTXLATOR.xlate_rec(expression, instance)
        elif select is None:
            return [_DICTXLATOR.xlate_rec(expression, i)
                    for i in self.contents]
        else:
            return [_DICTXLATOR.xlate_rec(expression, i)
                    for i in self.contents
                    if select.match(i)]

    @_supports_view(_translate)
    def translate(self, expression, instance=None):
        """Return list of translations for given expression.

        Parameters
        ----------
        instance : instance, optional
            Only translate given expression (returns a string).

        See also
        --------
        expand

        Examples
        --------
        >>> l = Launchgen()
        >>> l.params("a % 2 == 0",
        ...          a = range(4),
        ...          b = [ "foo", "bar" ])
        >>> l
        Launchgen([Instance({'a': 0, 'b': 'foo'}),
                   Instance({'a': 0, 'b': 'bar'}),
                   Instance({'a': 2, 'b': 'foo'}),
                   Instance({'a': 2, 'b': 'bar'})],
                  out='./out')

        >>> l.translate("@a@")
        ['0', '0', '2', '2']
        >>> l.translate("@a@-@b@", list(l.contents)[1])
        '0-bar'
        >>> l.translate("@a@-@b@", {'a': 3, 'b': 'baz'})
        '3-baz'

        """
        return self._translate(None, expression, instance)

    def _expand(self, select, expression, **kwargs):
        with_view = kwargs.pop("with_view", False)
        utils.assert_kwargs(kwargs)

        if select is None:
            filter_ = None
        else:
            filter_ = sciexp2.common.filter.Filter(select)

        expansions = _EXPANDER.expand(expression, self._contents,
                                      filter_=filter_,
                                      recursive=True,
                                      with_filter=with_view)
        if not with_view:
            expansions = expansions.keys()
        else:
            if filter_ is None:
                filter_ = sciexp2.common.filter.Filter()
            expansions = [(res.text, self.select(filter_ & res.filter))
                          for res in expansions.values()]
        return expansions

    @_supports_view(_expand)
    def expand(self, expression, **kwargs):
        """Return a list with the expansions of the given expression.

        Parameters
        ----------
        expression
            Expression to expand.
        with_view : bool, optional
            Return a list of expansion and `LaunchgenView` pairs (default is
            `False`).

        See also
        --------
        translate

        Examples
        --------
        >>> l = Launchgen()
        >>> l.params("a % 2 == 0",
        ...          a = range(4),
        ...          b = [ "foo", "bar" ])
        >>> l
        Launchgen([Instance({'a': 0, 'b': 'foo'}),
                   Instance({'a': 0, 'b': 'bar'}),
                   Instance({'a': 2, 'b': 'foo'}),
                   Instance({'a': 2, 'b': 'bar'})],
                  out='./out')

        >>> l.expand("@a@")
        ['0', '2']
        >>> l.expand("@a@", with_view=True)     #doctest: +NORMALIZE_WHITESPACE
        [('0', Launchgen([Instance({'a': 0, 'b': 'foo'}),
                          Instance({'a': 0, 'b': 'bar'})],
                         out='./out')),
         ('2', Launchgen([Instance({'a': 2, 'b': 'foo'}),
                          Instance({'a': 2, 'b': 'bar'})],
                         out='./out'))]
        >>> l.select("b == 'foo'").expand("@a@", with_view=True) #doctest: +NORMALIZE_WHITESPACE
        [('0', Launchgen([Instance({'a': 0, 'b': 'foo'})],
                         out='./out')),
         ('2', Launchgen([Instance({'a': 2, 'b': 'foo'})],
                         out='./out'))]

        """
        return self._expand(None, expression, **kwargs)

    def _generate_generic(self, select, from_expr, to_expr,
                          default_values, pre_values, filters, values,
                          post_generate):
        """Generate files using current contents and given values.

        Evaluates any deferred value before file generation.

        See also
        --------
        _do_generate

        """
        # defaults might be overriden only on some of the instances
        # separate those that the user has not specified from the rest
        default_values_maybe = dict(
            [(var, val)
             for var, val in six.iteritems(default_values)
             if var in self._contents or var in values])
        default_values = dict(
            [(var, val)
             for var, val in six.iteritems(default_values)
             if var not in default_values_maybe])

        # merge (non-defined) default_values into values
        values = dict(values)
        values.update(default_values)

        # merge contents and values
        contents, deferred = _do_params(select, self._contents, self._deferred,
                                        filters, values, False)

        # late evaluation of deferred values
        if len(deferred) is not 0:
            contents.cache_reset()
            for instance in deferred:
                for key in six.iterkeys(instance):
                    if not defer_evaluate(instance, key):
                        raise ValueError("Deferral of '%s' cannot "
                                         "be evaluated" % key)

        if select is not None:
            contents = sciexp2.common.instance.InstanceGroup(
                contents.select(select))

        if len(contents) == 0:
            return sciexp2.common.instance.InstanceGroup()
        else:
            return _do_generate(contents, from_expr, to_expr, self.out,
                                default_values, pre_values,
                                post_generate)

    def _generate(self, select, from_expr, to_expr, *filters, **values):
        def post_generate(path, instance, xlator):
            pass
        self._generate_generic(select, from_expr, to_expr,
                               {}, {}, filters, values,
                               post_generate)

    @_supports_view(_generate)
    def generate(self, from_expr, to_expr, *filters, **values):

        """Generate files by string translation.

        Much like `pack`, generates files from source templates, by way of
        translating all variable references into its values.

        For each expansion of `from_expr` with the current contents, uses the
        resulting instances to expand `to_expr`. Then uses `values` and the
        instance resulting of the `to_expr` expansion to translate the contents
        of the source.

        Parameters
        ----------
        from_expr, to_expr : expression
            Source/destination file expressions.
        filters : list of filters, optional
            Discard results not matching the given filters (default is to
            accept all).
        values : dict
            Additional variables to translate in destination files.

        Notes
        -----
        Arguments `filters` and `values` are interpreted as in `params` to
        (optionally) create some temporary contents that are then used to
        perform the actual file generation.

        If a single expansion of `to_expr` produces more than one matching
        instance, only the first one is used and noted with the message
        ``Skipping already generated file: <source> -> <destination>``.

        If it already exists, the destination file is only modified if its new
        contents are different from its current ones.

        See also
        --------
        params

        """
        return self._generate(None, from_expr, to_expr, *filters, **values)

    def _launcher(self, select, template, to_expr, *export, **values):
        tpl = templates.get(template)
        tpl_system = tpl.system

        jd = utils.get_path(values.pop("JD", "jobs.jd"))
        if not os.path.isabs(jd):
            jd = self.out + os.sep + jd
        out_abs = os.path.abspath(utils.get_path(self.out))
        jd = os.path.relpath(jd)
        jd_abs = os.path.abspath(jd)
        if jd_abs[:len(out_abs)] != out_abs:
            raise ValueError("Job descriptor file must be inside "
                             "the output directory")
        jd_base = os.path.dirname(jd_abs[len(out_abs)+1:])

        for reserved in set.union(tpl_system.defines(), ["LAUNCHER"]):
            if reserved in values or reserved in self._contents:
                raise ValueError("'%s' is a reserved variable" % reserved)

        values["LAUNCHER"] = to_expr

        if len(export) > 0:
            warnings.warn("Use of deprecated argument 'export' in method "
                          "'Launchgen.launcher'")
            exports = set(export)
        exports = set(values.pop("EXPORTS", []))

        depends = values.pop("DEPENDS", [])
        if isinstance(depends, six.string_types):
            raise TypeError("DEPENDS must be a sequence of strings")
        depends = set(str(d) for d in depends)
        depends.add("LAUNCHER")
        depends_unknown = [dep for dep in depends
                           if dep not in self._contents and dep not in values]
        if len(depends_unknown):
            raise ValueError("Unknown dependencies: %s" %
                             ",".join(depends_unknown))

        # TODO: should export used variables
        submit_args = values.pop("SUBMIT_ARGS", [])
        if isinstance(submit_args, six.string_types) or \
           not isinstance(submit_args, collections.Iterable):
            raise TypeError("Argument 'SUBMIT_ARGS' is not a list of strings")
        submit_args = tpl.submit_args + submit_args
        submit_args = [str(i) for i in submit_args]

        def post_generate(path, instance, xlator):
            tpl_system.post_generate(self.out, path, instance, xlator)
        group = self._generate_generic(select, tpl.template_path,
                                       to_expr,
                                       tpl.defaults, tpl.overrides,
                                       [], values,
                                       post_generate)

        if len(group) == 0:
            launcher.save(jd, jd_base, tpl.system, group, set(), depends,
                          submit_args)
        else:
            # select variables to export
            exports = set(tpl_system.assumes()) | depends | exports
            if len(exports - set(group.variables())) > 0:
                raise ValueError("Unknown variables to export: " +
                                 ", ".join(exports - set(group.variables())))

            # generate job descriptor file
            launcher.save(jd, jd_base, tpl.system, group, exports, depends,
                          submit_args)

    @_supports_view(_launcher)
    def launcher(self, template, to_expr, *export, **values):
        """Generate files from predefined templates.

        Performs the same work as `generate` with a predefined source file
        (identified by `template`) and additionally creates a job descriptor
        file in the output directory, which can be later used by `launcher`.

        The job descriptor will contain all variables used in `to_expr` (to
        identify the jobs by the parameters used to generate their files) and
        those listed in `export` (to provide further identification variables
        established by the user).

        Parameters
        ----------
        template : str
            Source template name
        to_expr : expression
            Destination job launcher file expression.
        export : list of str, optional
            Deprecated in favour of *EXPORTS* in argument `values`.
        values : dict
            Additional variables to translate in destination files.

        See also
        --------
        generate
        sciexp2.templates : Retrieve and manage launcher templates.

        Notes
        -----
        Variables *DONE* and *FAIL* are used as file paths by `launcher` to
        identify that a job has been run, and whether it has failed
        (respectively).

        If a variable in the current contents is also present in `values`, the
        one in `values` will overwrite the contents.

        The following variables are implicitly set:

        ========== =======================
        Name       Description
        ========== =======================
        *LAUNCHER* Expansion of `to_expr`.
        ========== =======================

        The following variables have a special meaning in argument `values`:

        ============= ============= ==========================================
        Name          Default value Description
        ============= ============= ==========================================
        *JD*          ``"jobs.jd"`` Job descriptor file to use with
                                    `launcher`. Not expanded.
        *EXPORTS*     []            Variable names to export.
        *DEPENDS*     ["LAUNCHER"]  Extra job dependencies. Not expanded.
                                    Automatically exported.
        *SUBMIT_ARGS* []            Extra arguments to job-submitting program.
        ============= ============= ==========================================

        *EXPORTS* is a list of variable names whose values will be exported to
        the job descriptor file. By exporting them, you can later use these
        variables as filters on *launcher* commands.

        *DEPENDS* is a list of variable names whose values point to files. If
        any of those files does not exist or is newer than *DONE*, the job will
        be considered for reexecution. Variable *LAUNCHER* is always implicitly
        added into the list. If the variable value for an instance contains the
        empty string, it is ignored as a dependency.

        You must export all variables (directly or indirectly) used in
        *SUBMIT_ARGS*.

        """
        return self._launcher(None, template, to_expr, *export, **values)


######################################################################
# Parameter recombination

def _get_product(contents_vars, params):
    """Compute what the given `params` are defining."""
    if len(params) > 0:
        var, vals = params[0]
        params = params[1:]
        for val in vals:
            for rest in _get_product(contents_vars, params):
                res = {var: val}
                res.update(rest[0])

                xlate = rest[1]
                deferred = rest[2]
                if isinstance(val, six.string_types) and \
                   len(set(varref.expr_get_vars("@", val)) & contents_vars) > 0:
                    xlate.add(var)
                elif isinstance(val, _DeferAtom) and\
                     len(val.deferred() & contents_vars) > 0:
                    deferred.add(var)

                yield (res, xlate, deferred)
    else:
        yield (dict(), set(), set())


def _do_recombine(base_instance,
                  product, product_in_xlate, product_in_deferred,
                  product_xlate, product_deferred,
                  new_deferred, new_deferred_vars):
    """Generate new instances using the given parameter products."""
    new_deferred.discard(base_instance)

    for prod in product:
        res = sciexp2.common.instance.Instance(base_instance)

        # resolve old value references on new values
        for var, val in six.iteritems(prod):
            if var in product_in_xlate:
                val = _DICTXLATOR.xlate_rec(val, base_instance,
                                            allow_unxlated=True)
            elif var in product_in_deferred:
                # TODO: if evaluation fails partially, it will only be
                #       reevaluated later on, when the original value is no
                #       longer available
                val = val.defer_evaluate(res)

            res[var] = val

        # resolve new references ASAP
        for var in product_xlate:
            res[var] = _DICTXLATOR.xlate_rec(res[var], res,
                                             allow_unxlated=True)
        still_deferred = False
        for var in product_deferred:
            if not defer_evaluate(res, var):
                still_deferred = True
                new_deferred_vars.add(var)
        if still_deferred:
            new_deferred.add(res)

        yield res


def _do_params(select, contents, deferred, filters, params, append):
    new_vars = set(contents.variables()) | set(six.iterkeys(params))
    filters = sciexp2.common.filter.Filter(*filters)
    filters.validate(new_vars)

    if len(params) == 0:
        return contents, deferred

    # normalize `params`
    for var, val in six.iteritems(params):
        if not isinstance(val, collections.Iterable) or \
           isinstance(val, six.string_types):
            val = [val]
        if not isinstance(val, collections.Sized):
            val = list(val)
        params[var] = val

    params_xlate = set([var for var, vals in six.iteritems(params)
                        if any(isinstance(val, six.string_types) and
                               len(set(varref.expr_get_vars("@", val))) > 0
                               for val in vals)])
    params_deferred = set([var for var, vals in six.iteritems(params)
                           if any(isinstance(val, _DeferAtom)
                                  for val in vals)])
    params_size = functools.reduce(operator.mul,
                                   (len(val) for val in six.itervalues(params)),
                                   1)

    if append:
        if select is None:
            new_size = params_size + params_size
            new_contents = contents
        else:
            new_size = len(contents) + params_size + params_size
            new_contents = sciexp2.common.instance.InstanceGroup()
    else:
        new_size = len(contents) * params_size + params_size
        new_contents = sciexp2.common.instance.InstanceGroup()
    new_deferred = set(deferred)

    with progress.get(new_size, msg="calculating parameters...") as progr:
        product = []
        product_in_xlate = set()
        product_in_deferred = set()

        # ensure `params` order across runs
        params_sorted = sorted(six.iteritems(params),
                               key=lambda i: i[0])

        for instance, xlate, deferred in _get_product(
                set(contents.variables()), params_sorted):
            product.append(instance)
            product_in_xlate |= xlate
            product_in_deferred |= deferred
            progr()

        def add_recombinations(instance):
            new_deferred_vars = set()
            for new_instance in _do_recombine(
                    instance, product,
                    product_in_xlate, product_in_deferred,
                    params_xlate, params_deferred,
                    new_deferred, new_deferred_vars):
                if len(new_deferred_vars) > 0:
                    filters.validate(new_vars - new_deferred_vars)
                if filters.match(new_instance):
                    new_contents.add(new_instance)
                progr()

        if append:
            if select is None:
                add_recombinations(sciexp2.common.instance.Instance())
            else:
                selected = False
                for instance in contents:
                    if select.match(instance):
                        selected = True
                        new_contents.add(instance)
                        progr()
                    else:
                        if selected:
                            selected = False
                            add_recombinations(
                                sciexp2.common.instance.Instance())
                        new_contents.add(instance)
                        progr()

        else:                           # not append
            if select is None:
                for instance in contents:
                    add_recombinations(instance)
            else:
                for instance in contents:
                    if not select.match(instance):
                        new_contents.add(instance)
                        progr(params_size)
                    else:
                        add_recombinations(instance)

    return new_contents, new_deferred


######################################################################
# File generation

def _do_generate(instances, from_expr, to_expr, to_dir,
                 default_values, pre_values,
                 post_generate):
    """Generate files using given contents and values.

    Parameters
    ----------
    instances : InstanceGroup
        Group to perform expansions and value substitutions.
    from_expr, to_expr : expression
        Source/destination file expressions.
    to_dir : str
        Destination directory (appended into `to_expr`).
    default_values : dict
        Substitutions to perform on certain instances if not present.
    pre_values : dict
        Initial substitutions to perform.
    post_generate : function
        Function to call after generating a file. Function receives the
        instance used to generate the file, and a path to the generated file.

    Returns
    -------
    InstanceGroup
        Group with an instance with fully expanded values for each
        destination file.

    See also
    --------
    _do_generate_from

    """
    res = sciexp2.common.instance.InstanceGroup()

    with progress.get(instances, msg="generating files...") as progr:
        # expand on source file
        if len(varref.expr_get_vars("@", from_expr, True)) == 0:
            # shortcut for a single source file (hopefully the usual path)
            for instance in _do_generate_from(
                    instances, from_expr,
                    to_expr, to_dir,
                    default_values, pre_values,
                    post_generate,
                    progr):
                res.add(instance)
        else:
            # not the optimal solution as a full expansion on 'from_expr' will
            # eat many unnecessary cicles, because the result will have to be
            # expanded again with 'to_expr'
            from_expansions = _EXPANDER.expand(from_expr, instances,
                                               grouped=True, recursive=True)
            for from_expand in six.itervalues(from_expansions):
                from_path = utils.get_path(from_expand.text)
                for instance in _do_generate_from(
                        from_expand.instances, from_path,
                        to_expr, to_dir,
                        default_values, pre_values,
                        post_generate,
                        progr):
                    res.add(instance)

    return res


def _do_generate_from(instances, from_path, to_expr, to_dir,
                      default_values, pre_values,
                      post_generate,
                      progr):
    """Generate files using given contents and values.

    Parameters
    ----------
    instances : InstanceGroup
        Group to perform expansions and value substitutions.
    from_path : str
        Source file path.
    to_expr : expression
        Destination file expression.
    to_dir : str
        Destination directory (appended into `to_expr`).
    default_values : dict
        Substitutions to perform on certain instances if not present.
    pre_values : dict
        Initial substitutions to perform.
    post_generate : function
        Function to call after generating a file. Function receives the
        instance used to generate the file, a path to the generated file, and
        the xlator used to generate the file.
    progr
        Progress reporter.

    Returns
    -------
    Instance generator
        Instances with fully expanded values for each destination file.

    See also
    --------
    _do_generate_file

    """
    with open(from_path) as from_file:
        from_contents = from_file.readlines()

    # apply pre_values and default_values once and for all
    new_from_contents = []
    for line in from_contents:
        line = _DICTXLATOR.xlate(line, pre_values, allow_unxlated=True)
        line = _DICTXLATOR.xlate_rec(line, default_values, allow_unxlated=True)
        new_from_contents.append(line)
    from_contents = new_from_contents

    # expand on destination file
    to_expansions = _EXPANDER.expand(to_expr, instances, recursive=True)
    for to_expand in six.itervalues(to_expansions):
        if len(to_expand.instances) > 1:
            progress.log(
                progress.LVL_INFO,
                "Skipping already generated file:",
                from_path, "->", to_expand.text, "\n")
            progress.log(
                progress.LVL_VERBOSE,
                "Conflicting instances:\n",
                to_expand.instances.get_index(0), "\n",
                to_expand.instances.get_index(1), "\n")

        to_path = utils.get_path(to_expand.text)
        if not os.path.isabs(to_path):
            to_path = os.path.join(to_dir, to_path)

        if not os.path.exists(to_path):
            with utils.get_file(to_path, "w") as to_file:
                progress.log(progress.LVL_DEBUG,
                             "Generating:", from_path, "->", to_path, "\n")
                res = _do_generate_file(to_expand.instances.get_index(0),
                                        from_contents, to_file, default_values)
                shutil.copymode(from_path, to_path)
                post_generate(to_path, res, _DICTXLATOR)
        else:
            with utils.get_tmp_file() as to_file:
                res = _do_generate_file(to_expand.instances.get_index(0),
                                        from_contents, to_file, default_values)
                to_file.flush()
                if utils.copy_path_maybe(to_file.name, to_path):
                    progress.log(progress.LVL_DEBUG,
                                 "Generating:", from_path, "->", to_path, "\n")
                    shutil.copymode(from_path, to_path)
                    post_generate(to_path, res, _DICTXLATOR)

        progr(len(to_expand.instances))
        yield res


def _do_generate_file(instance, from_contents, to_file, default_values):
    """Generate a file using given contents and values.

    Parameters
    ----------
    instance : Instance
        Instance to perform implicit translation with.
    from_contents : list(str)
        Untranslated source contents.
    to_file : file
        Destination file.
    default_values : dict
        Substitutions to perform on certain instances if not present.

    Returns
    -------
    Instance
        Instance with fully expanded values.

    """
    res = sciexp2.common.instance.Instance(instance)
    # make sure all values are fully expanded
    for var, val in six.iteritems(res):
        res[var] = _DICTXLATOR.xlate_rec(val, res)

    # if necessary, add defaults
    default_values_real = []
    for key, val in six.iteritems(default_values):
        if key not in res:
            res[key] = val
            default_values_real.append(key)
    # ensure added defaults are final
    # TODO: only necessary if the variable is going to be exported
    for key in default_values_real:
        res[key] = _DICTXLATOR.xlate_rec(res[key], res)

    # actually write file contents
    # (line-by-line to provide more readable errors)
    for line in from_contents:
        line = _DICTXLATOR.xlate(line, res)
        to_file.write(line)

    return res


######################################################################
# Views

class LaunchgenView (pp.Pretty):
    """Proxy to a subset of elements in a `Launchgen` instance.

    Notes
    -----
    Views follow the context manager protocol, which can be used to improve
    code readability:

    >>> l = Launchgen()
    >>> l.params(a=range(2), b=range(2))
    >>> with l.select("a != b") as v, v.select_inverse() as i:
    ...     v.params(c=1)
    ...     i.params(c=2)
    >>> l
    Launchgen([Instance({'a': 0, 'b': 0, 'c': 2}),
               Instance({'a': 0, 'b': 1, 'c': 1}),
               Instance({'a': 1, 'b': 0, 'c': 1}),
               Instance({'a': 1, 'b': 1, 'c': 2})],
              out='./out')

    """

    def __init__(self, launchgen, filter_base, filter_top,
                 **kwargs):
        assert len(kwargs) == 0
        self._launchgen = launchgen
        self._filters = [filter_base, filter_top]
        for filter_ in self._filters:
            filter_.validate(set(self._launchgen.contents.variables()))

    def _repr_pretty_(self, p, cycle):
        self._launchgen._do_repr_pretty_(self._as_filter(), p, cycle)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _as_filter(self):
        return sciexp2.common.filter.Filter(*self._filters)

    @property
    def contents(self):
        res = self._launchgen._contents.select(self._as_filter())
        return sciexp2.common.instance.InstanceGroup(res)

    def select(self, *filters):
        """Select a subset of this view.

        See also
        --------
        Launchgen.select

        """
        return LaunchgenView(self._launchgen,
                             self._as_filter(),
                             sciexp2.common.filter.Filter(*filters))

    def select_inverse(self):
        """Get the inverse selection of this view.

        If using nested views, this will invert the filter of this one, not all
        filters for parent views.

        See also
        --------
        Launchgen.select

        """
        return self._launchgen.select(
            sciexp2.common.filter.Filter(self._filters[0],
                                         "!(%s)" % self._filters[1]))

    def __getattr__(self, name):
        if name in ["_launchgen", "_filters"]:
            object.__getattribute__(self, name)
        else:
            attr = getattr(self._launchgen, name)
            if callable(attr):
                if attr.__name__ in VIEW_SUPPORTED:
                    attr = getattr(self._launchgen,
                                   VIEW_SUPPORTED[attr.__name__])

                def func(*args, **kwargs):
                    return attr(self._as_filter(), *args, **kwargs)
                return func
            else:
                return attr

    def __setattr__(self, name, value):
        if name in ["_launchgen", "_filters"]:
            object.__setattr__(self, name, value)
        else:
            attr = getattr(self._launchgen, name)
            assert not callable(attr)
            setattr(self._launchgen, name, value)


######################################################################
# Deferred evaluation

class _DeferAtom:
    """A piece in a deferred chain of operations."""

    def _check_deferred(self, maydefer, update=True):
        res = set()
        for i in maydefer:
            if isinstance(i, _DeferAtom):
                res |= i.deferred()
        if update:
            self._defer = res
        return res

    def deferred(self):
        """Return a set with deferred IDs."""
        return self._defer

    def __str__(self):
        raise ValueError("Tried to implicitly evaluate the deferral of: " +
                         ", ".join(self.deferred()))

    # definitions for implicit lookups

    def __add__(self, arg):
        return _DeferCall(self, "__add__", arg)

    def __radd__(self, arg):
        return _DeferCall(self, "__radd__", arg)

    def __sub__(self, arg):
        return _DeferCall(self, "__sub__", arg)

    def __rsub__(self, arg):
        return _DeferCall(self, "__rsub__", arg)

    def __mul__(self, arg):
        return _DeferCall(self, "__mul__", arg)

    def __rmul__(self, arg):
        return _DeferCall(self, "__rmul__", arg)

    def __div__(self, arg):
        return _DeferCall(self, "__div__", arg)

    def __rdiv__(self, arg):
        return _DeferCall(self, "__rdiv__", arg)

    def __mod__(self, arg):
        return _DeferCall(self, "__mod__", arg)

    def __rmod__(self, arg):
        return _DeferCall(self, "__rmod__", arg)


class _DeferInstance (_DeferAtom):
    """Deferred access to instance variable."""

    def __init__(self, name):
        self._name = name
        self._defer = set([name])

    def defer_evaluate(self, current, level=0):
        if self._name not in current:
            return self
        res = current[self._name]
        if isinstance(res, _DeferAtom):
            res = res.defer_evaluate(current, level+1)
        return res


class _DeferCall (_DeferAtom):
    """Deferred procedure call."""

    def __init__(self, obj, func, *args, **kwargs):
        assert isinstance(func, str) or callable(func)
        self._obj = obj
        self._func = func
        self._args = args
        self._kwargs = kwargs
        defer = list(self._args) + list(self._kwargs.values()) + [self._obj]
        self._check_deferred(defer)

    def defer_evaluate(self, current, level=0):
        # evaluate function arguments
        args = []
        for arg in self._args:
            if isinstance(arg, _DeferAtom):
                arg = arg.defer_evaluate(current, level+1)
            args.append(arg)

        kwargs = {}
        for key, val in six.iteritems(self._kwargs):
            if isinstance(val, _DeferAtom):
                val = val.defer_evaluate(current, level+1)
            kwargs[key] = val

        if len(self._check_deferred(args + list(kwargs.values()),
                                    False)) is not 0:
            return self

        # evaluate function
        if self._obj is None:
            return self._func(*args, **kwargs)
        else:
            obj = self._obj
            if isinstance(obj, _DeferAtom):
                obj = obj.defer_evaluate(current, level+1)
            if isinstance(obj, _DeferAtom):
                return self
            return getattr(obj, self._func)(*args, **kwargs)


def defer(arg, *args, **kwargs):
    """Return a deferrable value.

    Automatically detect the type of deferral through the type of `arg`:

      - variable access::

              params(P2 = defer("P1") * 2)
              params(P1 = [ 1, 2, 3, 4 ])

        Due to limitations in Python, non-numeric operations cannot be applied
        on the result of a call to `defer` (e.g., comparisons and boolean
        operators like `and` and `or`). In this case, a function call should be
        used.

      - function call::

              params(P2 = defer(math.log, defer("P1"), 2))
              params(P1 = [ 1, 2, 3, 4 ])

    Deferred values can be infinitely nested, such that these are not evaluated
    until all deferred values are available.

    """
    if isinstance(arg, str):
        assert len(args) is 0 and len(kwargs) is 0
        return _DeferInstance(arg)
    else:
        assert callable(arg)
        return _DeferCall(None, arg, *args, **kwargs)


def defer_evaluate(instance, key):
    """Replace ``instance[key]`` with the result of its evaluation.

    Returns whether the given key could be evaluated.

    """
    val = instance[key]
    if isinstance(val, _DeferAtom):
        val = val.defer_evaluate(instance)
        if isinstance(val, _DeferAtom):
            return False
        instance[key] = val
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Data manipulation abstractions."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2015, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import functools
import inspect
import numpy
import numpy as np                      # required for some numpy docstrings
from numpy.lib import recfunctions

from sciexp2.data import Data, data_array, _get_axis
from sciexp2.data import meta


def _wraps(name):
    def wrapper(func):
        doc = func.__doc__
        func.__doc__ = "`Data`-aware wrapper for `%s`" % (name)
        if doc is not None:
            func.__doc__ += "\n" + doc
        return func
    return wrapper

##################################################
# * Concatenate / append

_numpy_concatenate = numpy.concatenate


def _concatenate(arrays, axis):
    # calculate new np.array
    _, axis = _get_axis(axis, list(arrays))
    arr_res = _numpy_concatenate(arrays, axis=axis)
    assert isinstance(axis, int), "numpy.concatenate should only accept int axis"

    if not any(isinstance(array, Data) for array in arrays):
        return arr_res

    if axis is None:
        # flattened by numpy
        return data_array(arr_res, dtype=arr_res.dtype)

    # calculate new dimensions
    res_dims = []
    for didx in range(arrays[0].ndim):
        res_dim = None
        res_expr = None
        for array in arrays:
            if not isinstance(array, Data):
                continue
            a_dim = array.dims[didx]
            try:
                a_expr = a_dim.expression
            except meta.DimExpressionError:
                a_expr = None
            if res_expr is not None and a_expr is not None and res_expr != a_expr:
                raise ValueError(
                    "incompatible dimension expressions in dimension %d" % didx)
            if (res_dim is not None and a_dim is not None and
                list(res_dim) != list(a_dim) and didx != axis):
                raise ValueError(
                    "incompatible dimension contents in dimension %d" % didx)
            if res_dim is None:
                res_dim = a_dim
            if res_expr is None and a_expr is not None:
                res_dim = a_dim
                res_expr = a_expr

        res_dim = res_dim.copy()

        if didx == axis:
            res_dim_contents = []
            for array in arrays:
                if isinstance(array, Data):
                    res_dim_contents.extend(list(array.dims[didx]))
                else:
                    res_dim_contents.extend([None] * array.shape[didx])
                res_dim = meta.Dim(res_expr, res_dim_contents)

        res_dims.append(res_dim)

    return data_array(arr_res, dims=res_dims, dtype=arr_res.dtype)


@_wraps("numpy.concatenate")
def concatenate(arrays, axis=0):
    """
    For each dimension, uses the metadata of the first array with a non-null
    expression.

    """
    return _concatenate(arrays, axis=axis)


@_wraps("numpy.append")
def append(arr, values, axis=0):
    """
    For each dimension, uses the metadata of `values` if `arr` has no
    expression.

    """
    return _concatenate((arr, values), axis=axis)


######################################################################
# * Fields management

@_wraps("numpy.lib.recfunctions.append_fields")
def append_fields(base, names, data=None, dtypes=None, fill_value=-1):
    """
    The resulting format slightly deviates from NumpPy's to maintain the shape
    and metadata of `base`.

    """
    reference = [base]
    numpy_base = base.view(numpy.ndarray) \
                 if isinstance(base, Data) else base
    if isinstance(data, Data):
        reference += [data]
        numpy_data = data.view(numpy.ndarray).flat
    elif isinstance(data, collections.Iterable) and\
         not isinstance(data, numpy.ndarray):
        reference += data
        numpy_data = []
        for d in data:
            if isinstance(d, Data):
                d = d.view(numpy.ndarray)
            numpy_data.append(d)
    elif data is not None:
        reference += data
        numpy_data = data
    else:
        numpy_data = None

    res = recfunctions.append_fields(
        base=numpy_base, names=names, data=numpy_data,
        dtypes=dtypes, fill_value=fill_value,
        usemask=False, asrecarray=True)

    dims = None
    for arr in reference:
        if isinstance(arr, Data) and \
           reduce(int.__mul__, arr.shape) == reduce(int.__mul__, res.shape):
            dims = arr
            res = res.reshape(arr.shape)
            break
    if dims is None:
        raise ValueError("No input Data matches output shape (%s)" % res.shape)
    return data_array(res, dims=dims)


@_wraps("numpy.lib.recfunctions.drop_fields")
def drop_fields(*args, **kwargs):
    callargs = inspect.getcallargs(recfunctions.drop_fields, *args, **kwargs)
    res = recfunctions.drop_fields(**callargs)
    if (isinstance(callargs["base"], Data) and
        not callargs["usemask"] and
        not callargs["asrecarray"]):
        return data_array(res, dims=callargs["base"])
    else:
        return res


@_wraps("numpy.lib.recfunctions.merge_arrays")
def merge_arrays(*args, **kwargs):
    callargs = inspect.getcallargs(recfunctions.merge_arrays, *args, **kwargs)
    res = recfunctions.merge_arrays(**callargs)

    if callargs["usemask"] or callargs["asrecarray"]:
        return res

    dims = None
    for arr in callargs["seqarrays"]:
        if isinstance(arr, Data) and \
           reduce(int.__mul__, arr.shape) == reduce(int.__mul__, res.shape):
            dims = arr
            res = res.reshape(arr.shape)
            break
    if dims is None:
        raise ValueError("No input Data matches output shape (%s)" % res.shape)
    return data_array(res, dims=dims)


######################################################################
# * Miscellaneous

@_wraps("numpy.copy")
def copy(*args, **kwargs):
    callargs = inspect.getcallargs(numpy.copy, *args, **kwargs)
    try:
        func = callargs["a"].__class__.copy
    except AttributeError:
        func = numpy.copy
    return func(*args, **kwargs)


@_wraps("numpy.imag")
def imag(*args, **kwargs):
    callargs = inspect.getcallargs(numpy.imag, *args, **kwargs)
    try:
        func = callargs["val"].__class__.imag
    except AttributeError:
        func = numpy.imag
    return func(*args, **kwargs)


@_wraps("numpy.real")
def real(*args, **kwargs):
    callargs = inspect.getcallargs(numpy.real, *args, **kwargs)
    try:
        func = callargs["val"].__class__.real
    except AttributeError:
        func = numpy.real
    return func(*args, **kwargs)


@_wraps("numpy.ravel")
def ravel(*args, **kwargs):
    callargs = inspect.getcallargs(numpy.ravel, *args, **kwargs)
    try:
        func = callargs["a"].__class__.ravel
    except AttributeError:
        func = numpy.ravel
    return func(*args, **kwargs)


@_wraps("numpy.delete")
def delete(*args, **kwargs):
    callargs = inspect.getcallargs(numpy.delete, *args, **kwargs)

    _, axis = _get_axis(callargs["axis"],
                        [callargs["arr"]])
    callargs["axis"] = axis

    if callargs["arr"].ndim > 1 and axis is None:
        callargs["arr"] = callargs["arr"].flatten(order="A")

    res = numpy.delete(**callargs)

    arr = callargs["arr"]
    obj = callargs["obj"]
    axis = callargs["axis"]

    if not isinstance(arr, Data):
        return res

    daxis = axis if axis is not None else 0
    obj = arr.dims[daxis]._get_indexes(obj)

    dims = [(dim.expression, dim)
            for dim in arr._dims]
    if axis is None:
        axis = 0
    dims[axis] = (dims[axis][0],
                  list(numpy.delete(list(dims[axis][1]), obj)))
    dims = [meta.Dim(dim[0], dim[1])
            for dim in dims]
    return data_array(res, dims=dims)


__all__ = [
    "append", "concatenate",
    "append_fields", "drop_fields", "merge_arrays",
    "copy", "imag", "real", "ravel",
    "delete",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Simplified pretty-printing."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2011-2014, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import IPython.lib.pretty


class Pretty (object):
    """Provide repr/str through IPython's `pretty` module.

    Child classes must implement the `_repr_pretty_` method.

    """

    def pformat(self, p, cycle, name=None):
        """Return pretty representation.

        Automatically handles cycles and adds a proper group with the class
        name.

        This is designed to be used on a *with* statement.

        """
        if name is None:
            name = self.__class__.__name__

        if cycle:
            p.text("%s(...)" % name)
        else:
            begin = "%s(" % name
            return p.group(len(begin), begin, ")")

    def __repr__(self):
        return IPython.lib.pretty.pretty(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Errors raised during data extraction."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2014, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


class EmptyFileError (Exception):
    """File was unexpectedly empty."""

    def __init__(self, path):
        self.path = path

    def __str__(self):
        if self.path is None:
            return ""
        else:
            return self.path


class ExtractError (Exception):
    """Error extracting from file."""

    def __init__(self, path, exception):
        self.path = path
        self.exception = exception

    def __str__(self):
        msg = "Unexpected error while extracting source %s" % self.path
        msg += "\n  [was] "
        msg += "%s: %s" % (self.exception.__class__.__name__,
                           str(self.exception))
        return msg


class ProbeError (Exception):
    """Error probing from file."""

    def __init__(self, path, exception):
        self.path = path
        self.exception = exception

    def __str__(self):
        path = " %s" % self.path if self.path is not None else ""
        msg = "Unexpected error while probing source%s" % path
        msg += "\n  [was] "
        msg += "%s: %s" % (self.exception.__class__.__name__,
                           str(self.exception))
        return msg


__all__ = [
    "EmptyFileError", "ExtractError", "ProbeError",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Common utility classes and functions."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2015, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import re
import os
import shutil
import six
import subprocess
import tempfile
import warnings
import functools
import collections
import numpy as np
import weakref

from sciexp2.common import pp


# -----------------------------

def deprecated(func):
    """Decorator to mark routines as deprecated.

    It will result in a warning being emitted when the routine is used.
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn("Call to deprecated function '%s'" % func.__name__,
                      category=DeprecationWarning, stacklevel=2)
        return func(*args, **kwargs)
    return new_func


def assert_kwargs(kwargs):
    """Raise an exception if extra keys are present."""
    if len(kwargs) > 0:
        extra = "s" if len(kwargs) > 1 else ""
        raise TypeError("Unexpected argument%s: %s" % (extra,
                                                       ", ".join(kwargs)))


def text_realign(text, from_text, to_text):
    offset = len(from_text) - len(to_text)
    if offset > 0:
        lines = text.split("\n")
        text = [lines[0]] + [line[offset:]
                             for line in lines[1:]]
        text = "\n".join(text)
    elif offset < 0:
        lines = text.split("\n")
        text = [lines[0]] + [(" " * -offset) + line
                             for line in lines[1:]]
        text = "\n".join(text)
    return text.replace(from_text, to_text, 1)


# -----------------------------

def assert_dir(path):
    """Check that given directory exists, otherwise create it."""
    if path != "" and not os.path.exists(path):
        os.makedirs(path)


def assert_path(path):
    """Check that given path exists, otherwise create directories."""
    if not path.endswith(os.sep):
        path = os.path.dirname(path)
    assert_dir(path)


def get_path(path):
    path = os.path.expanduser(path)
    return os.path.expandvars(path)


def get_file(path, mod="w"):
    """Open the given file, creating any intermediate directory."""
    dir_path = os.path.dirname(path)
    assert_dir(dir_path)
    return open(path, mod)


def get_tmp_file(mode="w", delete=True):
    """Get a temporal file."""
    return tempfile.NamedTemporaryFile(mode=mode, delete=delete)


with open(os.devnull, "w") as _null:
    _have_rsync = subprocess.call(["which", "rsync"], stdout=_null)

if _have_rsync == 0:
    def copy_path_rsync(path_from, path_to, preserve=True, dereference=False):
        """Copy contents using rsync."""
        if os.path.isdir(path_from):
            path_from = path_from + os.sep
            assert_path(path_to)
        else:
            assert_path(os.path.dirname(path_to) + os.sep)

        args = "-rptgoD"
        if preserve:
            args += "t"
        if dereference:
            args += "l"
        else:
            args += "L"
        if subprocess.call(["rsync", args, path_from, path_to]) != 0:
            raise OSError("Error copying files: %s -> %s" % (
                path_from, path_to))

    def copy_path(*args, **kwargs):
        copy_path_rsync(*args, **kwargs)

else:
    def copy_path_shutil(path_from, path_to, preserve=True, dereference=False):
        """Copy contents using Python's shutil."""
        if os.path.isdir(path_from):
            # NOTE: will fail if destination already exists
            path_from = path_from + os.sep
            assert_path(path_to)
            shutil.copytree(path_from, path_to, symlinks=not dereference)

        else:
            assert_path(os.path.dirname(path_to) + os.sep)
            if os.path.islink(path_from):
                link_to = os.readlink(path_from)
                os.symlink(link_to, path_to)
            else:
                shutil.copy(path_from, path_to)
            if preserve:
                shutil.copymode(path_from, path_to)

    def copy_path(*args, **kwargs):
        copy_path_shutil(*args, **kwargs)


def copy_path_maybe(path_from, path_to):
    """Copy files only when contents differ."""
    if not os.path.exists(path_to) or os.path.isdir(path_from):
        copy_path(path_from, path_to)
        return True
    else:
        with open(path_from) as file_from,\
                open(path_to) as file_to:
            lines_from = file_from.readlines()
            lines_to = file_to.readlines()
        if lines_from == lines_to:
            return False
        else:
            copy_path(path_from, path_to)
            return True


# -----------------------------

def str2num(string):
    """Numeric value (``int`` or ``float``) of a string, if possible."""
    # TODO: use 'numpy.lib._iotools.StringConverter.upgrade'
    try:
        return int(string)
    except:
        try:
            return float(string)
        except:
            pass
    return string

# -----------------------------

# Set to True to see which instances are not memoized
SelectiveClassMemoize_DEBUG = False


class SelectiveClassMemoize (type):
    """Metaclass for selective memoization of object instantiation.

    This is useful for immutable objects that are costly to create and/or are
    created very often.

    Then, you must selectively use `memoize_new` to memoize those
    instances for which you wish to speedup instantiation::

       MyMemoizableClass.memoize_new(common set of construction arguments)
       MyMemoizableClass.memoize_new(another set set of common arguments)

    Thus, whenever you instantiate an object for the memoized class with the
    given set of object instantiation arguments, the memoized instance will be
    returned instead.

    .. warning::

       Mutation of memoized objects is strongly discouraged.

    """

    def __call__(cls, *args, **kwargs):
        mem = getattr(cls, "__SelectiveClassMemoize_mem")
        key = "%s" % repr((args, kwargs))
        if key in mem and cls is mem[key].__class__:
            return mem[key]
        if SelectiveClassMemoize_DEBUG:
            print("Not memoized:", cls, "::", key)
        return type.__call__(cls, *args, **kwargs)

    def memoize_new(cls, *args, **kwargs):
        """Create a memoized object with given parameters."""
        if not hasattr(cls, "__SelectiveClassMemoize_mem"):
            setattr(cls, "__SelectiveClassMemoize_mem", {})
        mem = getattr(cls, "__SelectiveClassMemoize_mem")
        key = "%s" % repr((args, kwargs))
        if not key in mem:
            mem[key] = type.__call__(cls, *args, **kwargs)


# -----------------------------

def _wraps(wrapped):
    return functools.wraps(wrapped=wrapped,
                           assigned=['__doc__'])


class ViewError (Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class OrderedSet (collections.MutableSet, pp.Pretty):
    """A mutable set preserving order of insertion.

    .. todo::

       All help should come from `~collections.MutableSet` instead of
       using `_wraps`.
    """

    @_wraps(collections.Container.__init__)
    def __init__(self, iterable=None, view_able=False):
        self._view_able = view_able
        if self._view_able:
            self._list = np.array([], dtype=object)
        else:
            self._list = []
        self._set_methods(False)
        self._base = None
        self._views = {}
        self._set = set()
        if iterable is not None:
            self |= iterable

    def view(self, index):
        assert self._view_able
        assert isinstance(index, slice)
        self._set_methods(True)
        res = OrderedSet([], True)
        res._list = self._list[index]
        for elem in res._list:
            res._set.add(elem)
        res._base = self
        self._views[id(res)] = weakref.ref(res)
        res._set_methods(True)
        return res

    def __del__(self):
        if self._base is not None:
            del self._base._views[id(self)]

    def _set_methods(self, view):
        if self._view_able:
            if view:
                self._append = self._append_array_view
                self._remove = self._remove_array_view
                self._pop = self._pop_array_view
            else:
                self._append = self._append_array
                self._remove = self._remove_array
                self._pop = self._pop_array
        else:
            assert not view
            self._append = self._append_list
            self._remove = self._remove_list
            self._pop = self._pop_list

    def _append_list(self, value):
        self._list.append(value)

    def _remove_list(self, value):
        if self._base is not None:
            self._base.remove(value)
        else:
            self._list.remove(value)
            for v in six.itervalues(self._views):
                v()._list.remove(value)
                v()._set.remove(value)

    def _pop_list(self, index):
        self._list.pop(index)

    def _append_array(self, value):
        self._list = np.append(self._list, value)

    def _remove_array(self, value):
        self._list = np.delete(self._list, np.where(self._list == value))

    def _pop_array(self, index):
        self._list = np.delete(self._list, index)

    def _append_array_view(self, value):
        raise ViewError("cannot append to a view")

    def _remove_array_view(self, value):
        raise ViewError("cannot remove from a view")

    def _pop_array_view(self, index):
        raise ViewError("cannot pop from a view")

    def _repr_pretty_(self, p, cycle):
        with self.pformat(p, cycle):
            p.pretty(list(self._list))

    def __repr__(self):
        return pp.Pretty.__repr__(self)

    def get_index(self, index):
        """Get item at the 'index'th position."""
        return self._list[index]

    def copy(self):
        return OrderedSet(self, self._view_able)

    def sorted(self, *args, **kwargs):
        """Same as `sort`, but returns a sorted copy."""
        res = self.copy()
        res.sort(*args, **kwargs)
        return res

    def sort(self, cmp=None, key=None, reverse=False):
        """Sort set in-place.

        Follows the same semantics of Python's built-in `sorted`.

        """
        if self._view_able:
            contents = list(self._list)
        else:
            contents = self._list

        contents.sort(cmp=cmp, key=key, reverse=reverse)

        if self._view_able:
            self._list[:] = contents

    # Container

    @_wraps(set.__contains__)
    def __contains__(self, key):
        return key in self._set

    # Sized

    @_wraps(set.__len__)
    def __len__(self):
        return len(self._list)

    # Iterable

    @_wraps(set.__iter__)
    def __iter__(self):
        return iter(self._list)

    # MutableSet

    def add(self, key):
        old_length = len(self._list)
        self._set.add(key)
        if len(self._set) != old_length:
            try:
                self._append(key)
            except ViewError:
                self._set.remove(key)
                raise
    add.__doc__ = collections.MutableSet.add.__doc__

    def discard(self, key):
        old_length = len(self._list)
        self._set.remove(key)
        if len(self._set) != old_length:
            try:
                self._remove(key)
            except ViewError:
                self._set.add(key)
                raise
    discard.__doc__ = collections.MutableSet.discard.__doc__
    discard.__doc__ += "\n\nThis operation has a cost of O(n)."

    @_wraps(collections.MutableSet.pop)
    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self._pop(-1) if last else self._pop(0)
        self._set.remove(key)
        return key

    # Pickling

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict["_append"]
        del odict["_remove"]
        del odict["_pop"]
        del odict["_base"]
        del odict["_views"]
        return odict

    def __setstate__(self, odict):
        self.__dict__.update(odict)
        self._base = None
        self._views = {}
        self._set_methods(False)


# -----------------------------

def _find_files(dir_path, cre):
    try:
        paths = iter(os.listdir(dir_path))
    except OSError:
        return
    for path in paths:
        cur_path = os.path.join(dir_path, path)
        if cre.match(cur_path):
            if os.path.isdir(cur_path):
                for sub_path in _find_files(cur_path, cre):
                    yield sub_path
            else:
                yield cur_path
        elif os.path.isdir(cur_path):
            for sub_path in _find_files(cur_path, cre):
                yield sub_path


def find_files(varre, file_expr, name=None, sort=True):
    """Find files matching a given expression.

    Returns an 'InstanceGroup' resulting from an instance extraction on each
    matching file.

    The argument `file_expr` can contain arbitrarily complex `regular
    expressions
    <http://docs.python.org/library/re.html#regular-expression-syntax>`_, and
    variables are extracted from:

        - variable references: ``@varname@``
        - named re groups: ``(?P<varname>regexp)``

    Parameters
    ----------
    varre
        A valid argument to `~sciexp2.common.varref.build_varre`.
    file_expr : str
        A regular expression for selecting files.
    name : str, optional
        On each resulting Instance, add a variable with the given name with the
        name of the file.
    sort : bool, optional
        Whether to sort the result according to the alphanumeric sorting of
        each of the variables in `file_expr`, in that specific order.

    Raises
    ------
    ValueError
        A variable name was specified in `name` and the expression `file_expr`
        contains such a variable name.

    """
    from . import varref
    import sciexp2.common.instance

    if not isinstance(file_expr, str):
        raise ValueError("Not an expression: " + file_expr)
    if name is not None and not isinstance(name, str):
        raise ValueError("'name' must be either None or an str")
    orig_file_expr = file_expr

    # convert into a regexp
    file_expr = varref.expr_to_regexp(varre, file_expr, contents=".+?")
    if name in varref.regexp_get_names(file_expr):
        raise ValueError(("Variable '%s'" % name) +
                         " is already present in the file expression:"
                         " '%s'" % orig_file_expr)

    # expand relative paths and other constructions
    file_expr = os.path.expandvars(file_expr)
    file_expr = os.path.expanduser(file_expr)
    file_expr = os.path.abspath(file_expr)

    # ensure there is no prefix match
    if not file_expr.endswith("$"):
        file_expr += "$"

    # get the initial directory to search
    start_dir = ""
    for part in file_expr.split(os.sep):
        path = os.sep.join([start_dir, part])
        if path[0:2] == os.sep * 2:
            path = path[1:]
        if os.path.isdir(path):
            start_dir = path
        else:
            break

    extractor = sciexp2.common.instance.InstanceExtractor(file_expr, varre)
    res = sciexp2.common.instance.InstanceGroup()

    file_cre = re.compile(file_expr)
    for file_path in _find_files(start_dir, file_cre):
        instance = extractor.extract(file_path)
        if instance is not None:
            # use numbers whenever possible (for later number-aware sorting)
            for key, val in six.iteritems(instance):
                instance[key] = str2num(val)

            if name:
                instance[name] = file_path

            if len(instance) > 0:
                res.add(instance)

    if sort:
        # sort result according to file sorting
        variables = varref.regexp_get_names(file_expr)
        res.sort(variables)

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Simple parallelization primitives.

By default, the amount of parallelism is controlled by the `PARALLELISM`
variable, interpreted according to `get_parallelism`.

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2013, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import contextlib
import multiprocessing
import multiprocessing.pool


#: Default amount of parallelism.
PARALLELISM = True


def get_parallelism(parallelism):
    """Compute the amount of available parallelism according to the system.

    Parameters
    ----------
    parallelism
        Requested parallelism.

    Notes
    -----
    The returned amount of parallelism depends on the value of `parallelism`:

    ================  ======================================================
    Value             Meaning
    ================  ======================================================
    `None`            Use `PARALLELISM` instead.
    `True`            Auto-detect as the number of cores in the system.
    positive integer  Use the given fixed amount.
    negative integer  Auto-detect as the number of cores in the system minus
                      the given value.
    ================  ======================================================

    """
    if parallelism is None:
        parallelism = PARALLELISM

    if parallelism is True:
        parallelism = multiprocessing.cpu_count()
    elif isinstance(parallelism, int):
        if parallelism == 0:
            raise ValueError("Invalid parallelism setting: %s" % parallelism)
        if parallelism < 0:
            parallelism = multiprocessing.cpu_count() + parallelism
    else:
        raise TypeError("Invalid parallelism setting: %s" % parallelism)

    return parallelism


#: Default amount of blocking.
BLOCKING = 50

#: Amount of blocking to use when the work length is unknown.
BLOCKING_UNKNOWN = 20

#: Maximum amount of blocking.
BLOCKING_MAX = 100


def get_blocking(blocking, work_len, parallelism=None):
    """Compute the amount of necessary blocking according to work length.

    Blocking establishes the amount of work items that each worker receives at
    a time. When using processes, this will reduce the costs of communication.

    Parameters
    ----------
    blocking
        Amount of blocking.
    work_len : int or None
        Work length (unknown if `None`).
    parallelism
        Argument to `get_parallelism`.

    Notes
    -----
    Argument `blocking` can mean different things depending on its value type:

    ================  ======================================================
    Value             Meaning
    ================  ======================================================
    `None`            Use `BLOCKING` instead.
    `True`            Evenly partition block with the amount of parallelism.
    positive integer  Use the given fixed amount.
    positive float    Use the given ratio of the given work elements, up to
                      `BLOCKING_MAX`.
    ================  ======================================================

    If the work size is unknown, `True` or a float argument will revert to
    `BLOCKING_UNKNOWN`.

    """
    if blocking is None:
        blocking = BLOCKING
    if blocking is True:
        if work_len is None:
            blocking = BLOCKING_UNKNOWN
        else:
            parallelism = get_parallelism(parallelism)
            blocking = work_len / parallelism
    elif isinstance(blocking, float):
        if work_len is None:
            blocking = BLOCKING_UNKNOWN
        else:
            blocking = min(work_len * blocking, BLOCKING_MAX)
    elif not isinstance(blocking, int):
        raise TypeError("Invalid blocking setting: %s" % blocking)
    return blocking


@contextlib.contextmanager
def configuration(**kwargs):
    """Context manager to temporarily override global parameters.

    Parameters
    ----------
    kwargs
        Temporary values to global parameters.

    Examples
    --------
    Temporarily override the default parallelism to all cores minus one and set
    blocking to 100:

    >>> def fadd (x) : return x + 1
    >>> def fmul (x) : return x * 2
    >>> with configuration(PARALLELISM = -1, BLOCKING = 100):
    ...    s = p_imap(range(10000), fadd)
    ...    res = p_imap(s, fmul)                       # doctest: +SKIP

    This is equivalent to:
    >>> s = p_imap(range(10000), fadd,
    ...            parallelism = -1, blocking = 100)   # doctest: +SKIP
    >>> res = p_imap(s, fmul,
    ...              parallelism = -1, blocking = 100) # doctest: +SKIP

    """
    variables = ["PARALLELISM", "BLOCKING", "BLOCKING_UNKNOWN", "BLOCKING_MAX"]
    backup = {}
    new = {}
    for var in variables:
        backup[var] = globals()[var]
        new[var] = kwargs.pop(var, backup[var])

    if len(kwargs) > 0:
        raise KeyError("Unknown globals: %s" % ", ".join(kwargs.keys()))

    globals().update(new)

    yield None

    globals().update(backup)


def _p_init(work, parallelism, blocking):
    if isinstance(work, collections.Sized):
        work_len = len(work)
    else:
        work_len = None

    parallelism = get_parallelism(parallelism)
    blocking = get_blocking(blocking, work_len, parallelism)
    pool = multiprocessing.Pool(processes=parallelism)
    return pool, blocking


def p_imap(work, func, parallelism=None, blocking=None):
    """Return a sequence with the result of mapping `func` to `work` in parallel.

    This function uses processes.

    Parameters
    ----------
    work : sequence
        Sequence of items to process in parallel.
    func : callable
        Function to apply on each element of `work`.
    parallelism
        Argument to `get_parallelism`.
    blocking
        Argument to `get_blocking`.

    See also
    --------
    multiprocessing.Pool, multiprocessing.Pool.imap

    """
    pool, blocking = _p_init(work, parallelism, blocking)
    res = iter(pool.imap(func, work, chunksize=blocking))
    pool.close()
    return res


def p_imap_unordered(work, func, parallelism=None, blocking=None):
    """Return a sequence with the result of mapping `func` to `work` in parallel.

    This function uses processes.

    Parameters
    ----------
    work : sequence
        Sequence of items to process in parallel.
    func : callable
        Function to apply on each element of `work`.
    parallelism
        Argument to `get_parallelism`.
    blocking
        Argument to `get_blocking`.

    See also
    --------
    multiprocessing.Pool, multiprocessing.Pool.imap_unordered

    """
    pool, blocking = _p_init(work, parallelism, blocking)
    res = iter(pool.imap_unordered(func, work, chunksize=blocking))
    pool.close()
    return res


def p_map(work, func, parallelism=None, blocking=None):
    """Return a list with the result of mapping `func` to `work` in parallel.

    This function uses processes.

    Parameters
    ----------
    work : sequence
        Sequence of items to process in parallel.
    func : callable
        Function to apply on each element of `work`.
    parallelism
        Argument to `get_parallelism`.
    blocking
        Argument to `get_blocking`.

    See also
    --------
    multiprocessing.Pool, multiprocessing.Pool.map

    """
    pool, blocking = _p_init(work, parallelism, blocking)
    res = pool.map(func, work, chunksize=blocking)
    pool.close()
    return res


def _t_init(work, parallelism, blocking):
    if isinstance(work, collections.Sized):
        work_len = len(work)
    else:
        work_len = None

    parallelism = get_parallelism(parallelism)
    blocking = get_blocking(blocking, work_len, parallelism)
    pool = multiprocessing.pool.ThreadPool(processes=parallelism)
    return pool, blocking


def t_imap(work, func, parallelism=None, blocking=None):
    """Return a sequence with the result of mapping `func` to `work` in parallel.

    This function uses threads.

    Parameters
    ----------
    work : sequence
        Sequence of items to process in parallel.
    func : callable
        Function to apply on each element of `work`.
    parallelism
        Argument to `get_parallelism`.
    blocking
        Argument to `get_blocking`.

    See also
    --------
    multiprocessing.pool.ThreadPool, multiprocessing.pool.ThreadPool.imap

    """
    pool, blocking = _t_init(work, parallelism, blocking)
    res = iter(pool.imap(func, work, chunksize=blocking))
    pool.close()
    return res


def t_imap_unordered(work, func, parallelism=None, blocking=None):
    """Return a sequence with the result of mapping `func` to `work` in parallel.

    This function uses threads.

    Parameters
    ----------
    work : sequence
        Sequence of items to process in parallel.
    func : callable
        Function to apply on each element of `work`.
    parallelism
        Argument to `get_parallelism`.
    blocking
        Argument to `get_blocking`.

    See also
    --------
    multiprocessing.pool.ThreadPool, multiprocessing.pool.ThreadPool.imap

    """
    pool, blocking = _t_init(work, parallelism, blocking)
    res = iter(pool.imap_unordered(func, work, chunksize=blocking))
    pool.close()
    return res


def t_map(work, func, parallelism=None, blocking=None):
    """Return a list with the result of mapping `func` to `work` in parallel.

    This function uses threads.

    Parameters
    ----------
    work : sequence
        Sequence of items to process in parallel.
    func : callable
        Function to apply on each element of `work`.
    parallelism
        Argument to `get_parallelism`.
    blocking
        Argument to `get_blocking`.

    See also
    --------
    multiprocessing.pool.ThreadPool, multiprocessing.pool.ThreadPool.map

    """
    pool, blocking = _t_init(work, parallelism, blocking)
    res = pool.map(func, work, chunksize=blocking)
    pool.close()
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Lazy function evaluation.

If possible, loads file from disk, otherwise computes the function and saves it
to disk.

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2013-2016, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import hashlib
import inspect
import os
import pickle
import six
import struct
import sys
import warnings

import sciexp2.data.io


##################################################
# Timestamp retrieval

def __timestamp_none(obj):
    return None


def __timestamp_find_files(obj):
    source = obj._Result__result()
    times = [None]
    for s in source:
        st = None
        sf = s["FILE"]
        if six.PY3:
            import io
            file = io.IOBase
        if isinstance(sf, six.string_types) and os.path.exists(sf):
            st = _timestamp_path(sf)
        elif isinstance(sf, file) and os.path.exists(sf):
            st = _timestamp_path(sf.name)
        else:
            st = None
        times.append(st)
    return max(times)


def _timestamp_path(path):
    if not os.path.exists(path):
        return None
    return os.stat(path).st_mtime


def __timestamp_sequence(elems):
    if len(elems) == 0:
        return None
    return max([_timestamp(elem) for elem in elems])


def _timestamp(elem):
    if isinstance(elem, tuple):
        return __timestamp_sequence(elem)
    elif isinstance(elem, set):
        return __timestamp_sequence(elem)
    elif isinstance(elem, dict):
        return max([__timestamp_sequence(elem.keys()),
                    __timestamp_sequence(elem.values())])
    elif isinstance(elem, list):
        return __timestamp_sequence(elem)
    elif isinstance(elem, Result):
        return elem._Result__timestamp()
    elif isinstance(elem, six.string_types):
        return None
    elif isinstance(elem, collections.Iterable):
        return __timestamp_sequence(elem)
    else:
        return None


##################################################
# Versioning control

def _version_write(fobj, version):
    fobj.write(struct.pack("I", len(version)))
    fobj.write(version)


def _version_read(fobj):
    len_fmt = "I"
    len_size = struct.calcsize(len_fmt)
    version_len = struct.unpack(len_fmt, fobj.read(len_size))[0]
    version = fobj.read(version_len)
    return version


def _version_skip(fobj):
    _version_read(fobj)


def __version_sequence(elems):
    return [_version(elem) for elem in elems]


def _version(elem):
    if isinstance(elem, tuple):
        res = __version_sequence(elem)
        return "tuple([%s])" % ", ".join(res)
    elif isinstance(elem, set):
        res = __version_sequence(elem)
        return "set([%s])" % ", ".join(res)
    elif isinstance(elem, dict):
        res = __version_sequence(six.iteritems(elem))
        return "{%s}" % ", ".join(res)
    elif isinstance(elem, list):
        res = __version_sequence(elem)
        return "[%s]" % ", ".join(res)
    elif isinstance(elem, Result):
        elem._Result__version()
        return elem._Result__version_str
    elif isinstance(elem, six.string_types):
        return repr(elem)
    elif isinstance(elem, collections.Iterable):
        res = __version_sequence(elem)
        return "[%s]" % ", ".join(res)
    elif inspect.isfunction(elem):
        source = inspect.getsource(elem)
        return hashlib.md5(source).hexdigest()
    else:
        return repr(elem)


##################################################
# Function-specific information

def __function_arg_is_lazy(obj):
    if isinstance(obj, dict):
        return any(__function_arg_is_lazy(key) or __function_arg_is_lazy(value)
                   for key, value in six.iteritems(obj))
    elif isinstance(obj, six.string_types):
        return False
    elif isinstance(obj, collections.Iterable):
        return any(__function_arg_is_lazy(i) for i in obj)
    else:
        return isinstance(obj, Result) or isinstance(obj, Method)


def __function_wrap_find_files(arg):
    return sciexp2.data.io._as_find_files(arg)[0]


def _function_parse(func, args, kwargs):
    fargs = inspect.getcallargs(func, *args, **kwargs)

    type_ = None
    timestamp = __timestamp_none

    if func == sciexp2.data.io.find_files or \
       func == __function_wrap_find_files:
        type_ = sciexp2.common.instance.InstanceGroup
        timestamp = __timestamp_find_files
    elif (func == sciexp2.data.io.extract_txt or
          func == sciexp2.data.io.extract_regex or
          func == sciexp2.data.io.extract_func):
        type_ = sciexp2.data.Data
        source = Result(__function_wrap_find_files,
                        [fargs["source"]], {})
        if "source" in kwargs:
            kwargs.pop("source")
            args = [source] + args
        else:
            args = list(args)
            args[0] = source

    return type_, args, kwargs, timestamp


##################################################
# Lazy function / method invocation

class LazySaveError (Exception):
    pass


def _lazy_result(obj):
    if isinstance(obj, Result):
        return obj._Result__result()
    elif isinstance(obj, dict):
        for key, value in six.iteritems(obj):
            obj[_lazy_result(key)] = _lazy_result(value)
        return obj
    elif isinstance(obj, tuple):
        return tuple([_lazy_result(i) for i in obj])
    elif isinstance(obj, list):
        return [_lazy_result(i) for i in obj]
    elif isinstance(obj, six.string_types):
        return obj
    else:
        return obj


def _lazy_outdated(obj):
    if isinstance(obj, Result):
        return obj._Result__outdated()
    elif isinstance(obj, dict):
        return any(_lazy_outdated(key) or _lazy_outdated(value)
                   for key, value in six.iteritems(obj))
    elif isinstance(obj, six.string_types):
        return False
    elif isinstance(obj, collections.Iterable):
        return any(_lazy_outdated(i) for i in obj)
    else:
        return False


class Result (object):

    def __init__(self, func, args, kwargs):
        if not callable(func):
            raise TypeError("Cannot lazily invoke a non-callable")
        self.__func = func

        self.__type, self.__args, self.__kwargs, self.__timestamp_func \
            = _function_parse(func, args, kwargs)

        self.__path = None
        self.__timestamp_num = None
        self.__version_str = None
        self.__version_num = None
        self.__res = None

    def __repr__(self):
        res = "lazy(%s" % self.__func
        if len(self.__args) > 0:
            res += ", " + ", ".join([repr(a) for a in self.__args])
        if len(self.__kwargs) > 0:
            res += ", " + ", ".join(["%s=%s" % (k, repr(v))
                                     for k, v in self.__kwargs.items()])
        res += ")"
        return res

    def __timestamp(self):
        if self.__timestamp_num is None:
            self.__timestamp_num = max(_timestamp(self.__args),
                                       _timestamp(self.__kwargs),
                                       self.__timestamp_func(self))
        return self.__timestamp_num

    def __version(self):
        if self.__version_num is None:
            # track function name and arguments
            self.__version_str = "%s.%s(%s)" % (
                self.__func.__module__, self.__func.__name__,
                ", ".join([_version(a) for a in self.__args] +
                          ["%s=%s" % (k, _version(v))
                           for k, v in six.iteritems(self.__kwargs)]))
            # track function source code
            if inspect.isfunction(self.__func):
                self.__version_str += ": {%s}" % inspect.getsource(self.__func)
            self.__version_num = hashlib.md5(self.__version_str).hexdigest()
        return self.__version_num

    def __outdated(self):
        if self.__path is not None and not os.path.exists(self.__path):
            return True
        res = _lazy_outdated(self.__args) or _lazy_outdated(self.__kwargs)
        if self.__path is not None and os.path.exists(self.__path):
            res |= self.__timestamp() >= _timestamp_path(self.__path)
            with open(self.__path, "rb") as fobj:
                res |= self.__version() != _version_read(fobj)
        return res

    ###########################################################
    # Result computation

    def __result_compute(self):
        self.__res = self.__func(*_lazy_result(self.__args),
                                 **_lazy_result(self.__kwargs))

    def __result_save(self):
        assert self.__path is not None
        version = self.__version()
        try:
            with open(self.__path, "wb") as fobj:
                _version_write(fobj, version)
                pickle.dump(self.__res, fobj)
        except Exception as e:
            os.unlink(self.__path)
            _, _, exc_traceback = sys.exc_info()
            new_e = LazySaveError("Cannot save: %s"
                                  "\n  [was] %s: %s" % (
                                      self, e.__class__.__name__, str(e)))
            six.reraise(new_e.__class__, new_e, exc_traceback)

    def __result_load(self):
        with open(self.__path, "rb") as fobj:
            _version_skip(fobj)
            self.__res = pickle.load(fobj)

    def __result(self):
        if self.__res is None:
            if self.__path is not None:
                if self.__outdated():
                    self.__result_compute()
                    self.__result_save()
                else:
                    self.__result_load()
            else:
                self.__result_compute()
        return self.__res

    ##################################################
    # Realization

    def checkpoint(self, path):
        self.__path = path
        return self

    def realize(self, path=None):
        self.__path = path
        return self.__result()

    def __getattr__(self, attr):
        return Method(self, attr)


def _method(obj, func, *args, **kwargs):
    func_obj = getattr(obj, func)
    if not callable(func_obj):
        raise TypeError("Non-callable lazy attribute: %s" % func)
    return func_obj(*args, **kwargs)


class Method:

    def __init__(self, object, func):
        self.__object = object
        self.__func = func

    def __call__(self, *args, **kwargs):
        return Result(_method,
                      [self.__object, self.__func] + list(args),
                      kwargs)

    def __getattr__(self, attr):
        raise Exception("Unsupported nested implicit lazy calls")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Progress indicators.

By default, messages are printed into `sys.stderr` (can be changed with
`set_file`), only if the provided level is greater than the one in `level`.

The available levels are: `LVL_NONE`, `LVL_MARK`, `LVL_PROGRESS` (the default),
`LVL_INFO`, `LVL_VERBOSE` and `LVL_DEBUG`.

In order to add the minimum amount of overhead, progress indicators are updated
at a maximum speed of `SPEED` (in seconds).

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2008-2014, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import collections
import datetime
import inspect
import itertools
import math
import numbers
import os
import six
import sys
import threading
import types


LVL_NONE = 0                      #: No progress indication.
LVL_MARK = 1                      #: Mark beginning/end of progress indication
LVL_PROGRESS = 2                  #: Show progress indication.
LVL_INFO = 3                      #: Show additional information.
LVL_VERBOSE = 4                   #: Be verbose when progressing.
LVL_DEBUG = 5                     #: Show debugging information.

_LEVEL = LVL_PROGRESS
_FILE = None
_INTERACTIVE = False


def set_file(fobj):
    global _FILE
    _FILE = fobj
    global _INTERACTIVE
    if hasattr(_FILE, "fileno") and os.isatty(fobj.fileno()):
        _INTERACTIVE = True
    else:
        _INTERACTIVE = False

set_file(sys.stderr)


def level(level=None):
    """Get/set the current progress indication level."""
    if level is not None:
        if level < LVL_NONE or LVL_DEBUG < level:
            raise ValueError("Invalid progress level %r" % level)
        global _LEVEL
        _LEVEL = level
    return _LEVEL


def _log(fmt, *args, **kwargs):
    with _print_lock:
        print(fmt, *args,  end="", file=_FILE, **kwargs)

_print_lock = threading.Lock()


def _logclean(length):
    global _LOGCLEAN
    _LOGCLEAN = (" " * length) + "\r"

_LOGCLEAN = ""


def log(level, fmt, *args, **kwargs):
    """Log message with given level."""
    if level > _LEVEL:
        return
    if _INTERACTIVE:
        _log(_LOGCLEAN)
    _log(fmt, *args, **kwargs)


def info(fmt, *args, **kwargs):
    log(LVL_INFO, fmt, *args, **kwargs)


def verbose(fmt, *args, **kwargs):
    log(LVL_VERBOSE, fmt, *args, **kwargs)


def debug(fmt, *args, **kwargs):
    log(LVL_DEBUG, fmt, *args, **kwargs)


#: Speed (in seconds) at which to update progress indicators.
SPEED = 0.2


def _block_get_length(count, length, delta):
    """Get a new block length.

    Smooths sudden changes by not allowing deltas beyond 50%.

    """
    new_length = count / (delta.total_seconds() / SPEED)
    new_length = int(math.ceil(new_length))
    if new_length > length:
        max_length = int(math.ceil(length * 1.5))
        return min(new_length, max_length)
    else:
        max_length = int(math.floor(length * 0.5))
        return max(new_length, max_length)


class Message:
    """A simple progress indicator.

    Parameters
    ----------
    msg : string, optional
        Message to prefix to the progress indicator.

    """

    def __init__(self, total=None, msg=""):
        self._interactive = _INTERACTIVE
        self._msg = " %s" % msg
        if self._interactive:
            self._msg += "\r"
            _log(self._msg)
        else:
            _log(self._msg + "\n")
        self._clean = False
        self._logclean = len(self._msg)
        _logclean(self._logclean)

    def __call__(self, increment=1):
        pass

    def _cleanup(self):
        if self._clean:
            return
        self._clean = True
        if self._interactive:
            length = len(self._msg)
            _log((" " * length) + "\r")
        else:
            _log(self._msg + " done.\n")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._cleanup()

    def __del__(self):
        """Delete the progress indicator from screen."""
        self._cleanup()


class Counter:
    """A counting progress indicator.

    Parameters
    ----------
    total : int
        Objective progress count.
    msg : string, optional
        Message to prefix to the progress indicator.

    """

    def __init__(self, total, msg=""):
        self._interactive = _INTERACTIVE
        self._total = total
        self._count = 0
        self._block_last_time = datetime.datetime.now()
        self._block_length = 1
        self._block_count = 0
        self._fmt_msg = " " + msg
        self._fmt = "%s  %%%%%dd / %%d" % (
            self._fmt_msg, len(str(self._total)))
        self._fmt = self._fmt % self._total
        self._clean = False
        if self._interactive:
            self._update = self._update_interactive
        else:
            self._update = self._update_noninteractive
        self._logclean = len(self._fmt % 0)
        _logclean(self._logclean)

    def __call__(self, increment=1):
        """Increment the progress indicator by the given number."""
        if level() < LVL_PROGRESS:
            return

        self._block_count += increment
        if self._block_count >= self._block_length:
            now = datetime.datetime.now()
            delta = now - self._block_last_time
            self._block_last_time = now
            self._block_length = _block_get_length(self._block_count,
                                                   self._block_length,
                                                   delta)
            self._count += self._block_count
            self._block_count = 0
            self._update()

    def _update_interactive(self):
        _log((self._fmt + "\r") % self._count)

    def _update_noninteractive(self):
        _log((self._fmt + "\n") % self._count)

    def _cleanup(self):
        if self._clean:
            return
        self._clean = True
        if self._interactive:
            length = len(self._fmt % 0)
            _log((" " * length) + "\r")
        else:
            _log(self._fmt_msg + " done.\n")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._cleanup()

    def __del__(self):
        """Delete the progress indicator from screen."""
        self._cleanup()


class Spinner:
    """A spinning (unbounded) progress indicator.

    Parameters
    ----------
    msg : string, optional
        Message to prefix to the progress indicator.

    """

    _SPINS = "-\\|/"

    def __init__(self, msg=""):
        self._interactive = _INTERACTIVE
        self._count = 0
        self._block_last_time = datetime.datetime.now()
        self._block_length = 1
        self._block_count = 0
        self._fmt_msg = " " + msg
        self._fmt = self._fmt_msg + " %s"
        self._clean = False
        if self._interactive:
            self._update = self._update_interactive
        else:
            self._update = self._update_noninteractive
        self._logclean = len(self._fmt % 0)
        _logclean(self._logclean)

    def __call__(self, increment=1):
        """Increment the progress indicator by the given number."""
        if level() < LVL_PROGRESS:
            return

        self._block_count += increment
        if self._block_count >= self._block_length:
            now = datetime.datetime.now()
            delta = now - self._block_last_time
            self._block_last_time = now
            self._block_length = _block_get_length(self._block_count,
                                                   self._block_length,
                                                   delta)
            self._count += 1
            self._count %= len(Spinner._SPINS)
            self._block_count = 0
            self._update()

    def _update_interactive(self):
        _log((self._fmt + "\r") % Spinner._SPINS[self._count])

    def _update_noninteractive(self):
        _log(self._fmt_msg + "\n")

    def _cleanup(self):
        if self._clean:
            return
        self._clean = True
        if self._interactive:
            length = len(self._fmt % " ")
            msg = (" " * length) + "\r"
            _log(msg)
        else:
            _log(self._fmt_msg + " done.\n")

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self._cleanup()

    def __del__(self):
        """Delete the progress indicator from screen."""
        self._cleanup()


class Null:
    """A "null" progress indicator.

    """

    def __init__(self, *args, **kwargs):
        self._logclean = 0
        _logclean(self._logclean)

    def __call__(self, *args, **kwargs):
        pass

    def _cleanup(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def __del__(self):
        pass


def _progressable_get_func_wrapper(func, other, progr):
    def wrapper(self, *args, **kwargs):
        res = func(other, *args, **kwargs)
        progr()
        return res
    return wrapper


def _progressable_get_iter_wrapper(func, other, progr):
    def wrapper(self, *args, **kwargs):
        otheriter = func(other, *args, **kwargs)

        class _ProgressableIterator (other.__class__):
            def __init__(self):
                pass

            def __iter__(self):
                return self

            def next(self):
                return self.__next__()

            def __next__(self):
                res = next(otheriter)
                progr()
                return res

            def __getattr__(self, attr):
                try:
                    return getattr(otheriter, attr)
                except:
                    return getattr(other, attr)

            def __setattr__(self, attr, val):
                setattr(otheriter, attr, val)

        return _ProgressableIterator()
    return wrapper


class Progressable:
    """Indicates objects returned by `progressable`."""
    pass


def progressable(other, progr, funcs=[], iters=[]):
    """Get a wrapper that invokes a progress indicator on some routines.

    Parameters
    ----------
    other
        Object to wrap.
    progr
        A progress indicator instance.
    funcs : list of str, optional
        List of routines in `other` to wrap.
    iters : list of str, optional
        List of routines in `other` that return an interator to wrap.

    Notes
    -----
    Argument `other` is also passed as the first argument to `get`, such that
    it can return a proper indicator.

    """
    assert len(funcs) > 0 or len(iters) > 0
    assert len(funcs) + len(iters) == len(set(funcs) | set(iters))

    class _Progressable (other.__class__, Progressable):
        def __init__(self):
            pass

        def __enter__(self):
            progr.__enter__()
            return self

        def __exit__(self, *args, **kwargs):
            progr.__exit__(*args, **kwargs)

        def __getattr__(self, attr):
            return getattr(other, attr)

        def __setattr__(self, attr, val):
            return other.__class__.__setattr__(other, attr, val)

    res = _Progressable()

    for func in itertools.chain(funcs, iters):
        if not hasattr(other, func):
            raise ValueError("Routine %s not found" % func)
        func_orig = getattr(other.__class__, func)
        if func in funcs:
            func_ = _progressable_get_func_wrapper(func_orig, other, progr)
        else:
            func_ = _progressable_get_iter_wrapper(func_orig, other, progr)
        setattr(_Progressable, func, func_)

    return res


def progressable_simple(contents, length, *args, **kwargs):
    """A simplified version of `progressable`.

    Creates a progress indicator from the given arguments.

    """
    if length is None:
        progr = get(contents, *args, **kwargs)
    else:
        progr = get(length, *args, **kwargs)

    if isinstance(contents, types.GeneratorType):
        class Wrapper (collections.Iterable, collections.Sized):
            def __iter__(self):
                return iter(contents)

            def __len__(self):
                return len(contents)

        res = Wrapper()
    else:
        res = contents

    return progressable(res, progr, iters=["__iter__"])


class Stack:
    """Stack of currently active progress indicators.

    Provides only static attributes and methods, in order to provide a
    program-wide stateful stack of progress indicators.
    """

    _STACK = []
    CURRENT = None
    """The currently active progress indicator."""

    @staticmethod
    def __call__(*args, **kwargs):
        """Call on the current (latest) progress indicator.

        Parameters
        ----------
        args, kwargs
            Arguments for ``__call__`` to the current progress indicator.
        """
        Stack.CURRENT(*args, **kwargs)

    @staticmethod
    def push(cls, *args, **kwargs):
        """Add a new progress indicator to the stack and make it current.

        Parameters
        ----------
        cls : callable
            Class object or function to construct the new progress indicator.
        args, kwargs
            Arguments to `cls`.
        """
        if not callable(cls):
            raise TypeError

        if level() < LVL_PROGRESS:
            obj = Null(*args, **kwargs)
        elif not _INTERACTIVE and cls != Null:
            obj = Message(*args, **kwargs)
        else:
            obj = cls(*args, **kwargs)

        Stack._STACK.append(obj)
        Stack.CURRENT = Stack._STACK[-1]
        return Stack.CURRENT

    @staticmethod
    def pop():
        """Remove the last progress indicator from the stack."""
        assert len(Stack._STACK) > 1
        res = Stack.CURRENT
        Stack._STACK.pop()
        Stack.CURRENT = Stack._STACK[-1]
        _logclean(Stack.CURRENT._logclean)
        return res

# Add a default null progress reporter
Stack.push(Null)


class UnpicklerStart:
    """Picklable proxy to a progress indicator.

    In order to provide a progress indicator when unpickling large objects, you
    can prefix the pickle stream with a `UnpicklerStart` instance and postfix
    it with its `UnpicklerStop` sister instance::

        def dump (self, file_obj):
            start = UnpicklerStart(len(self))
            pickle.dump((start, self, start.get_stop()), file_obj)

    When unpickled, the first will call `Stack.push` and the last will call
    `Stack.pop`, so that your object will be able to update the
    `~Stack.CURRENT` progress indicator by calling `Stack`.

    When pickled, it will also push a progress indicator for your code to use
    while pickling your large object.

    Parameters
    ----------
    args, kwargs
        Arguments for ``__init__`` to the progress indicator class.

    Notes
    -----
    The following keys in `kwargs` are reserved for UnpicklerStart
    itself:

    ================ ======== ===================================================
    Key              Type     Description
    ================ ======== ===================================================
    ``no_pickle``    boolean  Do not create a progress indicator while pickling
    ``cls_pickle``   callable Progress indicator class to use while pickling
    ``msg_pickle``   string   Message during pickling (otherwise use default)
    ``no_unpickle``  boolean  Do not create a progress indicator while unpickling
    ``cls_unpickle`` callable Progress indicator class to use while unpickling
    ``msg_unpickle`` string   Message during unpickling (otherwise use default)togg
    ================ ======== ===================================================

    By default both progress indicators are generated using some default
    messages.

    """

    def __init__(self, *args, **kwargs):
        no_pickle = kwargs.pop("no_pickle", False)
        cls_pickle = kwargs.pop("cls_pickle", get)
        msg_pickle = kwargs.pop("msg_pickle", "saving to file...")

        no_unpickle = kwargs.pop("no_unpickle", False)
        cls_unpickle = kwargs.pop("cls_unpickle", get)
        msg_unpickle = kwargs.pop("msg_unpickle", "loading from file...")

        if no_pickle:
            self._pickle = None
        else:
            pickle_kwargs = kwargs.copy()
            pickle_kwargs.setdefault("msg", msg_pickle)
            self._pickle = (cls_pickle, args, pickle_kwargs)

        if no_unpickle:
            self._unpickle = None
        else:
            unpickle_kwargs = kwargs.copy()
            unpickle_kwargs.setdefault("msg", msg_unpickle)
            self._unpickle = (cls_unpickle, args, unpickle_kwargs)

    def get_stop(self):
        """Get the sister `UnpicklerStop` object."""
        return UnpicklerStop(self._pickle is not None,
                             self._unpickle is not None)

    def __getstate__(self):
        contents = self.__dict__.copy()
        if self._pickle is not None:
            Stack.push(self._pickle[0],
                       *self._pickle[1],
                       **self._pickle[2])
        del contents["_pickle"]
        return contents

    def __setstate__(self, contents):
        self.__dict__.update(contents)
        if self._unpickle is not None:
            Stack.push(self._unpickle[0],
                       *self._unpickle[1],
                       **self._unpickle[2])


class UnpicklerStop:
    """Remove the progress indicators created by `UnpicklerStart`."""

    def __init__(self, do_pickle, do_unpickle):
        """
        Parameters
        ----------
        do_pickle, do_unpickle : boolean
            Remove a progress indicator after pickling / unpickling.
        """
        self._pickle = do_pickle
        self._unpickle = do_unpickle

    def __getstate__(self):
        contents = self.__dict__.copy()
        del contents["_pickle"]
        if self._pickle:
            progress = Stack.pop()
            progress._cleanup()
        return contents

    def __setstate__(self, contents):
        self.__dict__.update(contents)
        if self._unpickle:
            progress = Stack.pop()
            progress._cleanup()


def _get_with_level(cls, ignore_level, total_name, *args, **kwargs):
    if ignore_level:
        return cls
    lvl = level()
    if lvl >= LVL_PROGRESS:
        return cls(*args, **kwargs)
    if total_name is not None:
        assert len(args) == 0
        assert total_name in kwargs
        del kwargs[total_name]
    if lvl < LVL_MARK:
        return Null(*args, **kwargs)
    else:
        assert lvl == LVL_MARK
        return Message(*args, **kwargs)


def _get_ignore_level(*args, **kwargs):
    return get(*args, ignore_level=True, **kwargs)


def get(*args, **kwargs):
    """Return either a `Counter` or a `Spinner`.

    The type depends on wether the first argument is a number or an object
    whose length can be known.

    """
    ignore_lvl = kwargs.pop("ignore_level", False)

    try:
        cargs = inspect.getcallargs(Counter.__init__,
                                    *([None]+list(args)), **kwargs)
        del cargs["self"]
        first = cargs["total"]
    except:
        first = None
    if isinstance(first, six.string_types):
        # message in unnamed argument
        del cargs["total"]
        return _get_with_level(Spinner, ignore_lvl, None, **cargs)
    elif isinstance(first, numbers.Number):
        return _get_with_level(Counter, ignore_lvl, "total", **cargs)
    elif isinstance(first, collections.Sized):
        cargs["total"] = len(first)
        return _get_with_level(Counter, ignore_lvl, "total", **cargs)
    elif isinstance(first, collections.Iterable):
        # iterator
        del cargs["total"]
        return _get_with_level(Spinner, ignore_lvl, None, **cargs)
    else:
        return _get_with_level(Spinner, ignore_lvl, None, *args, **kwargs)


def get_pickle(*args, **kwargs):
    """Return an `UnpicklerStart` with either a `Counter` or a `Spinner`.

    The type depends on wether the first argument is a number or an object
    whose length can be known.

    """
    return UnpicklerStart(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Helper functions to handle data extraction sources."""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2014, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


from contextlib import contextmanager
import gzip
import six
import struct

from . import _errors


def get_file(arg):
    """Get a file-like object from `arg`."""
    if isinstance(arg, six.string_types):
        if arg.endswith(".gz"):
            return gzip.open(arg)
        else:
            return open(arg)
    else:
        return arg


def get_name(source, index):
    if isinstance(source, six.string_types):
        return source
    elif hasattr(source, "name"):
        return source.name
    elif index is None:
        return None
    else:
        return "number %d" % index


def get_size(source):
    with current_position(source):
        if isinstance(source, gzip.GzipFile):
            # little-endian unsigned integer in the last 4 bytes of the file
            source = source.myfileobj
            source.seek(0, 2)
            end_position = source.tell()
            assert end_position >= 4
            source.seek(-4, 2)
            return struct.unpack('<I', source.read(4))[0]
        else:
            source.seek(0, 2)
            return source.tell()


def is_empty(source):
    source_size = get_size(source)
    return source_size == 0


@contextmanager
def current_position(source):
    if isinstance(source, gzip.GzipFile):
        source = source.myfileobj
    source_pos = source.tell()
    yield
    source.seek(source_pos)


__all__ = [
    "get_file", "get_name", "get_size", "is_empty", "check_empty",
    "current_position",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""SciExp² version information."""

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2013-2016, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


__version_info__ = (1, 1, 3)
__version__ = ".".join([str(i) for i in __version_info__])


CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "Programming Language :: Python",
    "Topic :: Scientific/Engineering",
]


_description = "Scientific Experiment Exploration framework"

# taken from doc/introduction.rst
_long_description = """\
SciExp² (aka *SciExp square* or simply *SciExp2*) stands for *Scientific
Experiment Exploration*, which contains a comprehensive framework for easing
the workflow of creating, executing and evaluating experiments.

The driving idea behind the provided abstractions is that of the need for quick
and effortless *design-space exploration*. That is, the definition and
evaluation of experiments that are based on the permutation of different
parameters in the design space.

The framework is available in the form of Python modules which can be
integrated into your own applications, but also comes with some interfaces to
allow for quick writing without having to import a whole set of modules and
instantiate objects of their corresponding classes.
"""


NAME = "sciexp2"
VERSION = __version__
DESCRIPTION = _description
LONG_DESCRIPTION = _long_description
AUTHOR = __author__
AUTHOR_EMAIL = __email__
MAINTAINER = __maintainer__
MAINTAINER_EMAIL = __email__
URL = "https://projects.gso.ac.upc.edu/projects/sciexp2"
LICENSE = "GNU General Public License (GPL) version 3 or later"
PACKAGES = ["sciexp2", "sciexp2.common", "sciexp2.launchgen", "sciexp2.system",
            "sciexp2.data", "sciexp2.data.io"]
PACKAGE_DATA = {"sciexp2": ["templates/*.dsc", "templates/*.tpl"]}
SCRIPTS = ["launcher"]
REQUIRES = ["python (>=2.7)", "numpy", "ply", "ipython", "six"]
PROVIDES = ["sciexp2"]

INSTALL_REQUIRES = [req.replace("(", "").replace(")", "") for req in REQUIRES]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Common interface to manage jobs created with launchgen.

Each job is described with an `~sciexp2.common.instance.Instance`, as generated
by `~sciexp2.launchgen.Launchgen.launcher`.

You can integrate job management into your application by instantiating the
`Launcher` class.

"""

from __future__ import print_function
from __future__ import absolute_import

__author__ = "Lluís Vilanova"
__copyright__ = "Copyright 2009-2015, Lluís Vilanova"
__license__ = "GPL version 3 or later"

__maintainer__ = "Lluís Vilanova"
__email__ = "vilanova@ac.upc.edu"


import hashlib
import os
import stat
import struct
import time
from six.moves import cPickle as pickle
import subprocess

from .common import instance
from .common import progress
from .common import utils
from .common import varref
import sciexp2.system


_DICTXLATOR = varref.DictXlator("@")


class Launcher:
    """Common interface to manage jobs.

    The `types` argument to all methods identifies a set of job states to
    narrow the selection of jobs.

    The `filters` argument to all methods narrows the selection of jobs to
    those matching the given filters.

    """

    def __init__(self, base_path, system, group, depends, submit_args):
        """
        Parameters
        ----------
        base_path : str
            Path to the base directory created by
            `~sciexp2.launchgen.Launchgen.launcher`.
        system : str or System subclass
            Execution system.
        group : `~sciexp2.common.instance.InstanceGroup`
            Job descriptors.
        depends : sequence of str
            Variable names to which jobs depend.
        submit_args : sequence of str
            Extra arguments to the job-submitting program.

        See also
        --------
        sciexp2.system.get

        """
        if not os.path.isdir(base_path):
            raise IOError("Not a directory: " + base_path)

        if isinstance(system, str):
            name = system
            system = sciexp2.system.get(system)
            if system is None:
                raise ValueError("Not a launcher system name: " + name)
        if not issubclass(system, sciexp2.system.System):
            raise TypeError("Not a System subclass: " +
                            system.__class__.__name__)

        if not isinstance(group, instance.InstanceGroup):
            raise TypeError("Not an InstanceGroup")

        self._system = system(base_path, group, depends, submit_args)

    def parse_filters(self, *filters):
        value_filters = []
        filters_ = []
        for f in filters:
            try:
                f = sciexp2.common.filter.Filter(f)
                filters_.append(f)
            except sciexp2.common.filter.ParseError:
                path_abs = os.path.abspath(f)
                path = self._system.get_relative_path(
                    path_abs, self._system._base_abs)
                found = None
                for var in self.variables():
                    values = self._system._launchers[var]
                    if f in values:
                        found = "%s == %r" % (var, f)
                        break
                    if path in values:
                        found = "%s == %r" % (var, path)
                        break
                    if path_abs in values:
                        found = "%s == %r" % (var, path_abs)
                        break
                if found is None:
                    raise
                value_filters.append(found)
        res = sciexp2.common.filter.Filter(*filters_)
        if len(value_filters) > 0:
            res &= sciexp2.common.filter.Filter(" || ".join(value_filters))
        res.validate(self.variables())
        return res

    def variables(self):
        """Return a list with the available variables."""
        return list(self._system._launchers.variables())

    def values(self, *filters):
        """Return a list with the available variables."""
        filter_ = sciexp2.common.filter.Filter(*filters)
        return list(self._system._launchers.select(filter_))

    def summary(self, types, *filters):
        """Print a summary of the state of selected jobs."""
        if len(types) == 0:
            types = list(sciexp2.system.Job.STATES)

        parts = [(sciexp2.system.Job.STATE_LONG[t],
                  len(list(self._system.build([t], *filters))))
                 for t in types]
        max_name = max(len(t) for t, _ in parts)
        max_count = max(len(str(n)) for _, n in parts)
        fmt = "%%-%ds: %%%dd" % (max_name, max_count)
        for t, n in parts:
            print(fmt % (t, n))

    def state(self, types, *filters, **kwargs):
        """Print the states for the selected jobs."""
        expression = kwargs.pop("expression", None)
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))

        for job in self._system.build(types, *filters):
            state, name = job.state()
            if expression is not None:
                name = _DICTXLATOR.xlate_rec(expression, job)
            print("(%s) %s" % (sciexp2.system.Job.STATE_SHORT[state], name))

    def submit(self, types, *filters, **kwargs):
        """Submit selected jobs to execution.

        Calls `kill` before submitting a job if it's already running.

        """
        submit_args = kwargs.pop("submit_args", [])
        keep_going = kwargs.pop("keep_going", False)
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))
        jobs = list(self._system.build(types, *filters))
        exit_code = 0
        with progress.progressable_simple(
                jobs, None, msg="Submitting jobs...") as pjobs:
            for job in pjobs:
                if job["_STATE"] == sciexp2.system.Job.RUNNING:
                    job.kill()
                job_path = self._system.get_relative_path(job["LAUNCHER"])
                progress.log(progress.LVL_INFO, "Submitting", job_path, "\n")
                try:
                    job.submit(*submit_args)
                except subprocess.CalledProcessError as e:
                    exit_code = e.returncode
                    msg = "ERROR: Exited with code %d" % exit_code
                    if progress.level() < progress.LVL_INFO:
                        msg += " (%s)" % self._system.get_relative_path(job["LAUNCHER"])
                    progress.log(progress.LVL_NONE, msg, "\n")
                    if not keep_going:
                        break
        return exit_code

    def kill(self, types, *filters, **kwargs):
        """Kill the execution of selected jobs."""
        kill_args = kwargs.pop("kill_args", [])
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))
        for job in self._system.build(types, *filters):
            if job["_STATE"] == sciexp2.system.Job.RUNNING:
                job.kill(*kill_args)

    def files(self, types, *filters, **kwargs):
        """List files matching an expression."""
        expression = kwargs.pop("expression")
        not_expanded = kwargs.pop("not_expanded", False)
        if len(kwargs) > 0:
            raise ValueError("Unknown arguments: " + " ".join(kwargs.keys()))

        expr_path = expression
        if not os.path.isabs(expr_path):
            expr_path = os.path.join(self._system._base_abs, expr_path)

        expr_files = list(f["__FILE"]
                          for f in utils.find_files("@", expr_path,
                                                   name="__FILE"))

        job_files = set()
        for job in self._system.build(types, *filters):
            path = _DICTXLATOR.xlate_rec(expression, job)
            abs_path = path
            if not os.path.isabs(abs_path):
                abs_path = os.path.join(self._system._base_abs, abs_path)
            if not os.path.exists(abs_path):
                continue
            job_files.add(abs_path)

        if not_expanded:
            res = set(expr_files) - job_files
        else:
            res = set(expr_files) & job_files
        for i in sorted(res, cmp=lambda i1, i2: cmp(expr_files.index(i1),
                                              expr_files.index(i2))):
            print(i)


def _header():
    return "#!/usr/bin/env launcher\n"


def _magic():
    version = "launcher"
    return hashlib.md5(version).hexdigest()


class LauncherLoadError (Exception):
    """Could not load given file."""

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "Not a job descriptor file: %s" % self.path


def save(file_name, base_to_file, system, group, export, depends, submit_args):
    """Save an InstanceGroup as a job descriptor into a file.

    Parameters
    ----------
    file_name : str
        Path to destination file.
    base_to_file : str
        Relative path from some base directory to the directory containing
        `file_name`.
    system : `~sciexp2.system.System`
        Class of the execution system to use.
    group : `~sciexp2.common.instance.InstanceGroup`
        Job descriptors.
    export : set
        Variable names to export into the description file.
    depends : sequence of str
        Variable names to which jobs depend.
    submit_args : sequence of str
        Extra arguments to the job-submitting program.

    See also
    --------
    load, sciexp2.system.get

    """
    assert system is not None
    if not issubclass(system, sciexp2.system.System):
        system = sciexp2.system.get(system)

    if len(group) > 0:
        for assumed in system.assumes():
            if assumed not in group:
                raise ValueError("'%s' must be defined" % assumed)

    file_dir = os.path.dirname(file_name)
    if not os.path.exists(file_dir):
        os.makedirs(file_dir)
    with open(file_name, "wb") as file_obj:
        file_obj.write(_header())

        version = _magic()
        file_obj.write(struct.pack("I", len(version)))
        file_obj.write(version)

        pickle.dump((base_to_file, system.name, depends, submit_args),
                    file_obj, -1)
        group.dump(file_obj, export)
    fmode = stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IXOTH
    os.chmod(file_name, fmode)


def load(file_name):
    """Load a job description from a file.

    Parameters
    ----------
    file_name : str
        Path to source file.

    Returns
    -------
    sciexp2.system.System
        Instance of the execution system with the job descriptions.

    Raises
    ------
    LauncherLoadError
        The given file cannot be loaded as a `Launcher`.

    See also
    --------
    save
    """
    with open(file_name, "rb") as file_obj:
        header = file_obj.readline()
        if header != _header():
            raise LauncherLoadError(file_name)
        version_size = struct.unpack("I", file_obj.read(struct.calcsize("I")))
        version_size = version_size[0]
        version = file_obj.read(version_size)
        if version != _magic():
            raise LauncherLoadError(file_name)

        base_to_file, system_name, depends, submit_args = pickle.load(file_obj)
        group = instance.InstanceGroup.load(file_obj)

    jd_base = os.path.dirname(os.path.abspath(file_name))
    if base_to_file != "":
        base = jd_base[:-len(base_to_file)]
    else:
        base = jd_base
    base = os.path.relpath(base)
    return Launcher(base, system_name, group, depends, submit_args)
